import asyncio
import shutil
from datetime import datetime
from pathlib import Path

import pytest

import esgpull.plugin
from esgpull.esgpull import Esgpull
from esgpull.models import File, FileStatus
from esgpull.plugin import (
    Event,
    PluginManager,
)


@pytest.fixture
def plugin_dir(tmp_path: str):
    """Create a temporary plugin directory for testing"""
    temp_dir = Path(tmp_path) / "plugins"
    temp_dir.mkdir(parents=True)
    return temp_dir


@pytest.fixture
def config_path(plugin_dir: Path) -> Path:
    """Create a temporary config file path"""
    return plugin_dir / "plugins.toml"


@pytest.fixture
def plugin_manager(config_path: Path) -> PluginManager:
    """Create a plugin manager with a temporary config file"""
    pm = PluginManager(config_path)
    pm.enabled = True
    esgpull.plugin._plugin_manager = pm
    return pm


@pytest.fixture
def assets_path() -> Path:
    """Path to the assets directory containing sample plugins."""
    return Path(__file__).parent / "assets"


def copy_plugin(plugin_name: str, plugin_dir: Path, assets_path: Path):
    plugin_file = f"{plugin_name}.py"
    shutil.copy(assets_path / plugin_file, plugin_dir / plugin_file)
    return plugin_name


@pytest.fixture
def sample_plugin(plugin_dir: Path, assets_path: Path) -> str:
    return copy_plugin("sample_plugin", plugin_dir, assets_path)


@pytest.fixture
def incompatible_plugin(plugin_dir: Path, assets_path: Path) -> str:
    return copy_plugin("incompatible_plugin", plugin_dir, assets_path)


@pytest.fixture
def priority_test_plugin(plugin_dir: Path, assets_path: Path) -> str:
    return copy_plugin("priority_test_plugin", plugin_dir, assets_path)


@pytest.fixture
def error_plugin(plugin_dir: Path, assets_path: Path) -> str:
    return copy_plugin("error_plugin", plugin_dir, assets_path)


def test_plugin_discovery_and_loading(
    plugin_dir: Path,
    plugin_manager: PluginManager,
    sample_plugin: str,
):
    """Test that plugin discovery finds and loads plugins correctly"""
    plugin_manager.discover_plugins(plugin_dir, load_all=True)
    assert sample_plugin in plugin_manager.plugins

    # Check the plugin attributes
    plugin = plugin_manager.plugins[sample_plugin]
    assert plugin.name == sample_plugin
    assert plugin.min_version == "0.8.0"
    assert plugin.max_version is None

    # Verify config class was found
    assert hasattr(plugin.module, "Config")
    assert plugin.module.Config.log_level == "INFO"
    assert plugin.module.Config.max_retries == 3

    # Check that handlers are registered for the right events
    assert any(
        h.plugin_name == sample_plugin and h.event == Event.file_complete
        for h in plugin_manager._handlers_by_event[Event.file_complete]
    )
    assert any(
        h.plugin_name == sample_plugin and h.event == Event.file_error
        for h in plugin_manager._handlers_by_event[Event.file_error]
    )
    assert any(
        h.plugin_name == sample_plugin and h.event == Event.dataset_complete
        for h in plugin_manager._handlers_by_event[Event.dataset_complete]
    )

    # Verify priorities
    file_complete_handler = next(
        h
        for h in plugin_manager._handlers_by_event[Event.file_complete]
        if h.plugin_name == sample_plugin
    )
    assert file_complete_handler.priority == "normal"

    file_error_handler = next(
        h
        for h in plugin_manager._handlers_by_event[Event.file_error]
        if h.plugin_name == sample_plugin
    )
    assert file_error_handler.priority == "high"

    dataset_complete_handler = next(
        h
        for h in plugin_manager._handlers_by_event[Event.dataset_complete]
        if h.plugin_name == sample_plugin
    )
    assert dataset_complete_handler.priority == "low"


def test_discovery_only_enabled(
    plugin_dir: Path,
    plugin_manager: PluginManager,
    sample_plugin: str,
):
    """Test that plugin discovery finds and loads plugins correctly"""
    plugin_manager.discover_plugins(plugin_dir)
    assert len(plugin_manager.plugins) == 0
    plugin_manager.enable_plugin(sample_plugin)
    plugin_manager.discover_plugins(plugin_dir)
    assert len(plugin_manager.plugins) == 1


def test_plugin_enable_disable(plugin_dir, plugin_manager, sample_plugin):
    """Test enabling and disabling plugins"""
    plugin_manager.discover_plugins(plugin_dir, load_all=True)
    plugin_manager.enable_plugin(sample_plugin)
    assert sample_plugin in plugin_manager.config.enabled
    assert plugin_manager.is_plugin_enabled(sample_plugin) is True

    plugin_manager.disable_plugin(sample_plugin)
    assert sample_plugin in plugin_manager.config.disabled
    assert sample_plugin not in plugin_manager.config.enabled
    assert plugin_manager.is_plugin_enabled(sample_plugin) is False


def test_triggered_plugin(plugin_dir, plugin_manager, sample_plugin):
    plugin_manager.discover_plugins(plugin_dir, load_all=True)
    # Create a test File object
    test_file = File(
        file_id="enable_disable_test",
        dataset_id="test_dataset_id",
        master_id="test_master_id",
        url="http://example.com/test.nc",
        version="1.0",
        filename="enable_disable_test.nc",
        local_path="/path/to/test",
        data_node="test_node",
        checksum="12345",
        checksum_type="md5",
        size=1000,
        status=FileStatus.Done,
    )
    destination = Path()

    plugin_manager.enable_plugin(sample_plugin)
    plugin_manager.plugins[sample_plugin].module.calls["file_complete"] = 0
    plugin_manager.trigger_event(
        Event.file_complete,
        file=test_file,
        destination=destination,
        start_time=datetime.now(),
        end_time=datetime.now(),
    )
    assert (
        plugin_manager.plugins[sample_plugin].module.calls["file_complete"]
        == 1
    )

    plugin_manager.disable_plugin(sample_plugin)
    plugin_manager.plugins[sample_plugin].module.calls["file_complete"] = 0
    plugin_manager.trigger_event(
        Event.file_complete,
        file=test_file,
        destination=destination,
        start_time=datetime.now(),
        end_time=datetime.now(),
    )
    assert (
        plugin_manager.plugins[sample_plugin].module.calls["file_complete"]
        == 0
    )


def test_version_compatibility(
    plugin_dir: Path,
    plugin_manager: PluginManager,
    sample_plugin: str,
    incompatible_plugin: str,
):
    """Test version compatibility checks for plugins"""
    plugin_manager.discover_plugins(plugin_dir, load_all=True)
    assert sample_plugin in plugin_manager.plugins
    assert incompatible_plugin not in plugin_manager.plugins

    # Check that only the compatible plugin's handlers are registered
    assert any(
        h.plugin_name == sample_plugin
        for h in plugin_manager._handlers_by_event[Event.file_complete]
    )
    assert not any(
        h.plugin_name == incompatible_plugin
        for h in plugin_manager._handlers_by_event[Event.file_complete]
    )


def test_priority_order(plugin_dir, plugin_manager, priority_test_plugin):
    """Test that handlers execute in priority order"""
    plugin_manager.enable_plugin(priority_test_plugin)
    plugin_manager.discover_plugins(plugin_dir)
    assert priority_test_plugin in plugin_manager.plugins
    plugin = plugin_manager.plugins[priority_test_plugin]
    plugin.module.execution_order = []

    test_file = File(
        file_id="priority_test_file",
        dataset_id="test_dataset_id",
        master_id="test_master_id",
        url="http://example.com/test.nc",
        version="1.0",
        filename="priority_test.nc",
        local_path="/path/to/test",
        data_node="test_node",
        checksum="12345",
        checksum_type="md5",
        size=1000,
        status=FileStatus.Done,
    )
    destination = Path()
    plugin_manager.trigger_event(
        Event.file_complete,
        file=test_file,
        destination=destination,
        start_time=datetime.now(),
        end_time=datetime.now(),
    )
    assert plugin.module.execution_order == ["high", "normal", "low"]


def test_plugin_config_set(plugin_dir, plugin_manager, sample_plugin):
    """Test plugin set config value"""
    plugin_manager.enable_plugin(sample_plugin)
    plugin_manager.discover_plugins(plugin_dir)
    assert sample_plugin in plugin_manager.plugins

    plugin = plugin_manager.plugins[sample_plugin]
    assert plugin.module.Config.log_level == "INFO"
    assert plugin.module.Config.notification_method == "console"
    assert plugin.module.Config.max_retries == 3

    assert sample_plugin in plugin_manager.config.plugins
    plugin_config = plugin_manager.config.plugins[sample_plugin]
    assert plugin_config["log_level"] == "INFO"
    assert plugin_config["notification_method"] == "console"
    assert plugin_config["max_retries"] == 3

    # Modify a config value
    plugin_manager.set_plugin_config(sample_plugin, "log_level", "DEBUG")
    assert plugin_config["log_level"] == "DEBUG"
    assert plugin.module.Config.log_level == "DEBUG"
    assert plugin.module.Config.notification_method == "console"
    assert plugin.module.Config.max_retries == 3


def test_plugin_config_load(plugin_dir, plugin_manager, sample_plugin):
    """Test plugin load config after settings values"""
    plugin_manager.enable_plugin(sample_plugin)
    plugin_manager.discover_plugins(plugin_dir)
    plugin_manager.set_plugin_config(sample_plugin, "log_level", "DEBUG")
    plugin_manager.set_plugin_config(sample_plugin, "max_retries", "150")
    plugin_manager.write_config(generate_full_config=True)

    # Create new plugin manager with same config file
    new_pm = PluginManager(plugin_manager.config_path)
    new_pm.enabled = True
    new_pm.load_config()
    new_pm.discover_plugins(plugin_dir)
    plugin_config = new_pm.config.plugins[sample_plugin]
    assert plugin_config["log_level"] == "DEBUG"
    assert plugin_config["notification_method"] == "console"
    assert plugin_config["max_retries"] == 150
    plugin = new_pm.plugins[sample_plugin]
    assert plugin.module.Config.log_level == "DEBUG"
    assert plugin.module.Config.notification_method == "console"
    assert plugin.module.Config.max_retries == 150


def test_error_isolation(
    plugin_dir, plugin_manager, sample_plugin, error_plugin
):
    """Test that errors in one plugin don't affect others"""
    plugin_manager.enable_plugin(sample_plugin)
    plugin_manager.enable_plugin(error_plugin)
    plugin_manager.discover_plugins(plugin_dir)
    assert sample_plugin in plugin_manager.plugins
    assert error_plugin in plugin_manager.plugins
    error_plugin_calls = plugin_manager.plugins[error_plugin].module.calls
    sample_plugin_calls = plugin_manager.plugins[sample_plugin].module.calls
    error_plugin_calls["file_complete"] = 0
    sample_plugin_calls["file_complete"] = 0

    # Create a test File object
    test_file = File(
        file_id="error_isolation_test",
        dataset_id="test_dataset_id",
        master_id="test_master_id",
        url="http://example.com/test.nc",
        version="1.0",
        filename="error_isolation_test.nc",
        local_path="/path/to/test",
        data_node="test_node",
        checksum="12345",
        checksum_type="md5",
        size=1000,
        status=FileStatus.Done,
    )
    destination = Path()

    # All handlers are correctly called even if one raises
    plugin_manager.trigger_event(
        Event.file_complete,
        file=test_file,
        destination=destination,
        start_time=datetime.now(),
        end_time=datetime.now(),
    )
    assert error_plugin_calls["file_complete"] == 1
    assert sample_plugin_calls["file_complete"] == 1

    with pytest.raises(ValueError):
        # Trigger the file_complete event
        # The error_plugin has high priority and will run first and raise an exception
        plugin_manager.trigger_event(
            Event.file_complete,
            file=test_file,
            destination=destination,
            start_time=datetime.now(),
            end_time=datetime.now(),
            reraise=True,
        )
    assert error_plugin_calls["file_complete"] == 2
    assert sample_plugin_calls["file_complete"] == 1


def test_download_event_workflow(assets_path, tmp_path):
    """Test plugin integration with the download workflow"""

    # Create an Esgpull instance with install=True to set up the directory
    install_path = tmp_path / "esgpull"
    esg = Esgpull(path=install_path, install=True)

    # Copy the sample plugin to the plugins directory
    esg.config.paths.plugins.mkdir(parents=True, exist_ok=True)
    plugin_name = "sample_plugin"
    plugin_file = f"{plugin_name}.py"
    shutil.copy(
        assets_path / plugin_file, esg.config.paths.plugins / plugin_file
    )

    # Create an Esgpull instance
    esg = Esgpull(install_path)
    esg.plugin_manager.enabled = True
    esg.plugin_manager.enable_plugin(plugin_name)
    esg.plugin_manager.discover_plugins(esg.config.paths.plugins)
    assert plugin_name in esg.plugin_manager.plugins
    module = esg.plugin_manager.plugins[plugin_name].module

    # Create a File object for our test file
    test_file = File(
        file_id="test_file",
        dataset_id="esgpull",
        master_id="license",
        url="https://raw.githubusercontent.com/ESGF/esgf-download/refs/heads/main/LICENSE",
        version="1",
        filename="license",
        local_path="test/path",
        data_node="github",
        checksum="954a863c65831c7c639140ee6208cb2d8d974175e7fc41d4edf9eb4161c6ba07",
        checksum_type="SHA256",
        size=1540,
        status=FileStatus.Queued,
    )
    test_file.compute_sha()
    module.calls["file_complete"] = 0
    module.calls["file_error"] = 0
    coro = esg.download([test_file], show_progress=False, use_db=False)
    downloaded, errors = asyncio.run(coro)
    assert len(downloaded) == 1
    assert len(errors) == 0
    assert module.calls["file_complete"] == 1
    assert module.calls["file_error"] == 0

    # Create a File object with a nonexistent URL
    failing_file = File(
        file_id="failing_file",
        dataset_id="test_dataset_id",
        master_id="test_master_id",
        url="https://nonexistent.path/to/file.nc",
        version="1.0",
        filename="failing.nc",
        local_path="test/failing",
        data_node="test_node",
        checksum="12345",
        checksum_type="SHA256",
        size=1000,
        status=FileStatus.Queued,
    )
    failing_file.compute_sha()
    module.calls["file_complete"] = 0
    module.calls["file_error"] = 0
    coro = esg.download([failing_file], show_progress=False, use_db=False)
    downloaded, errors = asyncio.run(coro)
    assert len(downloaded) == 0
    assert len(errors) == 1
    assert module.calls["file_complete"] == 0
    assert module.calls["file_error"] == 1



from datetime import datetime

import pytest

from esgpull.utils import format_date, format_date_iso, parse_date


def test_parse_date():
    fmt = "%Y-%m-%d"
    expected = datetime(year=2022, month=1, day=1)
    assert parse_date("2022-01-01", fmt=fmt) == expected
    assert parse_date(datetime(year=2022, month=1, day=1), fmt=fmt) == expected
    with pytest.raises(ValueError):
        parse_date("20220101", fmt=fmt)
    with pytest.raises(ValueError):
        parse_date(20220101, fmt=fmt)


def test_format_date():
    fmt = "%Y-%m-%d"
    expected = "2022-01-01"
    assert format_date("2022-01-01", fmt=fmt) == expected
    assert (
        format_date(datetime(year=2022, month=1, day=1), fmt=fmt) == expected
    )
    with pytest.raises(ValueError):
        format_date("20220101", fmt=fmt)
    with pytest.raises(ValueError):
        format_date(20220101, fmt=fmt)


def test_format_date_iso():
    expected = "2022-01-01T00:00:00Z"
    assert format_date_iso("2022-01-01") == expected
    assert format_date_iso(datetime(year=2022, month=1, day=1)) == expected
    with pytest.raises(ValueError):
        format_date_iso("20220101")
    with pytest.raises(ValueError):
        format_date_iso(20220101)



"""update tables

Revision ID: 0.7.3
Revises: 0.7.2
Create Date: 2024-09-20 12:22:11.989996

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0.7.3'
down_revision = '0.7.2'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



import os
from configparser import ConfigParser
from pathlib import Path

import click
from click.exceptions import Abort, Exit
from rich.table import Table

from esgpull import Esgpull
from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import init_esgpull
from esgpull.config import Config
from esgpull.exceptions import (
    InvalidInstallPath,
    UnknownInstallName,
    UnregisteredInstallPath,
)
from esgpull.install_config import InstallConfig
from esgpull.tui import TempUI, Verbosity


def get_synda_db_path(sdt_home: str | None = None) -> Path | None:
    if sdt_home is None:
        sdt_home = os.getenv("SDT_HOME")
    if sdt_home is not None:
        sdt_path = Path(sdt_home).expanduser().resolve()
        conf_path = sdt_path / "conf" / "sdt.conf"
        conf = ConfigParser()
        conf.read(conf_path)
        path = Path(conf["core"]["db_path"]) / "sdt.db"
        if path.is_file():
            db_path = path
        else:
            db_path = None
    else:
        db_path = None
    return db_path


@click.group()
def self():
    """
    Manage esgpull installations / import synda database
    """
    ...


@self.command()
@args.path
@opts.name
@opts.verbosity
def install(
    path: Path | None,
    name: str | None,
    verbosity: Verbosity,
):
    with TempUI.logging():
        TempUI.rule("[b green]esgpull[/] installation")
        if path is None:
            default = str(InstallConfig.default)
            path = Path(TempUI.prompt("Install location", default=default))
            idx = InstallConfig.index(path=path)
            if idx > -1:
                TempUI.print(
                    (
                        f"\n:stop_sign: {path} is already installed:\n"
                        f"  {InstallConfig.installs[idx]}\n\n"
                        f"{InstallConfig.activate_msg(idx)}"
                    ),
                    err=True,
                )
                raise Exit(1)
            if name is None:
                name = TempUI.prompt("Name (optional)") or None
        idx = InstallConfig.add(path, name)
        InstallConfig.choose(idx=idx)
        path = InstallConfig.installs[idx].path
        if path.is_dir():
            TempUI.print(f"Using existing install at {path}")
        else:
            TempUI.print(f"Creating install directory and files at {path}")
        InstallConfig.write()
        TempUI.print(f"Install config added to {InstallConfig.path}")
        esg = Esgpull(verbosity=verbosity, install=True)
    with esg.ui.logging("init", onraise=Abort):
        # with esg.ui.spinner("Fetching facets"):
        #     if esg.fetch_facets(update=False):
        #         esg.ui.print(":+1: Facets are initialised.")
        #     else:
        #         esg.ui.print(":+1: Facets were already initialised.")
        sdt_home = os.getenv("SDT_HOME")
        if sdt_home is not None:
            db_name = get_synda_db_path(sdt_home) or ""
            msg = (
                f"Found existing synda installation at {sdt_home}\n"
                "You can import its database by running:\n"
                f"$ esgpull self import_synda {db_name}"
            )
            esg.ui.print(msg)


@self.command()
@args.path
@opts.name
def activate(
    path: Path | None,
    name: str | None,
):
    with TempUI.logging():
        idx = InstallConfig.index(path=path, name=name)
        if idx < 0:
            if path is not None and InstallConfig.index(name=path.name) > -1:
                raise ValueError(
                    f"{InstallConfig.fullpath(path)} is not installed, "
                    "did you mean this?\n\n"
                    f"$ esgpull self activate --name {path}"
                )
            raise InvalidInstallPath(path=path)
        TempUI.print(InstallConfig.activate_needs_eval(idx))


@self.command()
@args.path
@opts.name
def choose(
    path: Path | None,
    name: str | None,
):
    with TempUI.logging():
        if path is None and name is None:
            table = Table(
                title="Install locations",
                title_style="bold",
                title_justify="left",
                show_header=False,
                show_footer=False,
                show_edge=False,
                box=None,
                padding=(0, 1),
            )
            table.add_column(style="b yellow")
            table.add_column(style="magenta")
            table.add_column(style="b green")
            for i, inst in enumerate(InstallConfig.installs):
                table.add_row(
                    "*" if i == InstallConfig.current_idx else "",
                    str(inst.path),
                    inst.name or "",
                )
            TempUI.print(table)
            raise Exit(0)
        InstallConfig.choose(path=path, name=name)
        if InstallConfig.current is None:
            if name is not None:
                raise UnknownInstallName(name)
            elif path is not None:
                raise UnregisteredInstallPath(path)
        else:
            InstallConfig.write()


@self.command()
def reset():
    with TempUI.logging():
        if InstallConfig.current is not None:
            idx = InstallConfig.current_idx
            path = InstallConfig.current.path
            InstallConfig.choose()
            InstallConfig.write()
            TempUI.print(
                f"Install location is not set anymore as {path}\n"
                # "To set it back to its previous location, run:\n"
                # f"$ esgpull self choose {name}"
            )
            TempUI.print(InstallConfig.activate_msg(idx))
            raise Exit(0)
        else:
            TempUI.print(":stop_sign: No install found.")
            TempUI.print("To install esgpull, run:\n")
            TempUI.print("$ esgpull self install")
            raise Exit(1)


@self.command()
def delete():
    with TempUI.logging():
        if InstallConfig.current is None:
            msg = (
                "None\n"
                "Please choose an existing install before trying to delete it."
                "\n\n$ esgpull self choose ..."
            )
            raise UnregisteredInstallPath(msg)
        else:
            path = InstallConfig.current.path
            TempUI.print(f"You are going to delete: {path}")
            choice = TempUI.prompt(f"Please enter {path.name!r} to continue")
            if choice == path.name:
                TempUI.print(f"Deleting {path} from config...")
                TempUI.print("To remove all files from this install, run:\n")
                config = Config.load(path=path)
                for p in config.paths.values():
                    if not p.is_relative_to(path):
                        TempUI.print(f"$ rm -rf {p}")
                TempUI.print(f"$ rm -rf {path}")
                InstallConfig.remove_current()
                InstallConfig.write()
            else:
                raise Abort


@self.command()
@args.path
@opts.verbosity
def import_synda(
    path: Path | None,
    verbosity: Verbosity,
):
    esg = init_esgpull(verbosity)
    with esg.ui.logging("import_synda", onraise=Abort):
        if path is None:
            sdt_home = os.getenv("SDT_HOME")
            prompt_title = "Enter synda database location"
            if sdt_home is not None:
                esg.ui.print(
                    "Found existing synda installation at"
                    f" SDT_HOME={sdt_home}"
                )
                default = str(get_synda_db_path(sdt_home))
                path = Path(esg.ui.prompt(prompt_title, default=str(default)))
            else:
                path = Path(esg.ui.prompt(prompt_title))
        else:
            path = path.expanduser().resolve()
        if not path.is_file():
            raise FileNotFoundError(path)
        nb_imported = esg.import_synda(url=path, track=True, ask=True)
        esg.ui.print(f"Imported {nb_imported} new files from {path}")



import pytest

from esgpull import Esgpull
from esgpull.models import Query


def test_insert_default_query(root):
    esg = Esgpull(root, install=True)
    default_query = Query(selection=dict(project="IPSL"))
    default_query.compute_sha()
    esg.graph.add(default_query)
    esg.config.api.default_query_id = default_query.sha
    queries = [
        Query(selection=dict(variable="tas")),
        Query(selection=dict(institution_id="toto")),
        Query(selection=dict(time_frequency="day")),
        Query(selection=dict(variable="pr")),
    ]
    for q in queries:
        q.compute_sha()
    queries.append(
        Query(require=queries[-1].sha, selection=dict(member_id="r1i1p1f1"))
    )
    queries[-1].compute_sha()
    queries.append(
        Query(require=queries[-1].sha, selection=dict(table_id="Amon"))
    )
    queries[-1].compute_sha()
    new_queries = esg.insert_default_query(*queries)
    esg.graph.add(*new_queries)
    assert len(esg.graph.queries) == 7
    for query in queries:
        with pytest.raises(KeyError):
            esg.graph.get(query.sha)
    for i, query in enumerate(new_queries):
        assert esg.graph.get(query.sha) == query
        if i in range(4):
            assert query.require == default_query.sha
        elif i == 4:
            assert query.require == new_queries[3].sha
        elif i == 5:
            assert query.require == new_queries[4].sha



# Plugin for testing execution priority order

from esgpull.plugin import Event, on

# Track execution order for testing
execution_order = []


@on(Event.file_complete, priority="high")
def handle_file_complete_high(file, destination, start_time, end_time, logger):
    """High priority handler for file_complete event"""
    logger.info(f"High priority handler: {file.filename}")
    execution_order.append("high")


@on(Event.file_complete, priority="normal")
def handle_file_complete_normal(
    file, destination, start_time, end_time, logger
):
    """Normal priority handler for file_complete event"""
    logger.info(f"Normal priority handler: {file.filename}")
    execution_order.append("normal")


@on(Event.file_complete, priority="low")
def handle_file_complete_low(file, destination, start_time, end_time, logger):
    """Low priority handler for file_complete event"""
    logger.info(f"Low priority handler: {file.filename}")
    execution_order.append("low")



"""update tables

Revision ID: 0.9.5
Revises: 0.9.4
Create Date: 2026-02-04 15:52:02.010431

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0.9.5'
down_revision = '0.9.4'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from collections import OrderedDict
from textwrap import dedent

import click
from click.exceptions import Abort, BadOptionUsage

from esgpull import Esgpull
from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import extract_subdict, init_esgpull, totable
from esgpull.models import Dataset, File, FileStatus
from esgpull.plugin import Event, emit
from esgpull.tui import Verbosity
from esgpull.version import __version__


@click.group()
def plugins():
    """Manage plugins."""
    pass


def check_enabled(esg: Esgpull):
    if not esg.plugin_manager.enabled:
        esg.ui.print("Plugin system is [red]disabled[/]")
        esg.ui.print("To enable it, run:")
        esg.ui.print("    esgpull config plugins.enabled true")
        raise Abort


@plugins.command("ls")
@click.option(
    "--json", "json_output", is_flag=True, help="Output in JSON format"
)
@opts.verbosity
def list_plugins(verbosity: Verbosity, json_output: bool = False):
    """List all available plugins and their status."""
    esg = init_esgpull(verbosity=verbosity)

    with esg.ui.logging("plugins", onraise=Abort):
        check_enabled(esg)
        esg.plugin_manager.discover_plugins(
            esg.config.paths.plugins, load_all=True
        )
        if json_output:
            # Format for JSON output
            result = {}
            for name, p in esg.plugin_manager.plugins.items():
                # Get handlers by event type
                handlers = {}
                for h in p.handlers:
                    event_type = h.event.value
                    if event_type not in handlers:
                        handlers[event_type] = []
                    handlers[event_type].append(
                        {
                            "function": h.func.__name__,
                            "priority": h.priority,
                        }
                    )

                result[name] = {
                    "enabled": esg.plugin_manager.is_plugin_enabled(name),
                    "handlers": handlers,
                    "min_version": p.min_version,
                    "max_version": p.max_version,
                }

            esg.ui.print(result, json=True)
        else:
            # Format for human-readable output
            if not esg.plugin_manager.plugins:
                esg.ui.print("No plugins found.")
                return

            # Prepare table data
            table_data = []
            for name, p in esg.plugin_manager.plugins.items():
                enabled = esg.plugin_manager.is_plugin_enabled(name)
                status_circle = "🟢" if enabled else "🔴"
                plugin_name = f"{status_circle} {name}"

                # Collect event handlers with their details
                handler_rows = []
                events_by_type = {}
                for h in p.handlers:
                    event_type = h.event.value
                    if event_type not in events_by_type:
                        events_by_type[event_type] = []
                    events_by_type[event_type].append(h.func.__name__)

                # Create handler rows with event type and function names
                for event_type, handler_names in events_by_type.items():
                    for handler_name in handler_names:
                        handler_rows.append(
                            {"event": event_type, "function": handler_name}
                        )

                # First row with plugin info and first handler
                if handler_rows:
                    first_handler = handler_rows[0]
                    first_row = OrderedDict(
                        [
                            ("plugin", plugin_name),
                            ("event", first_handler["event"]),
                            ("function", first_handler["function"]),
                        ]
                    )
                    table_data.append(first_row)

                    # Additional rows for remaining handlers
                    for handler in handler_rows[1:]:
                        additional_row = OrderedDict(
                            [
                                ("plugin", ""),
                                ("event", handler["event"]),
                                ("function", handler["function"]),
                            ]
                        )
                        table_data.append(additional_row)
                else:
                    # Plugin with no handlers
                    row = OrderedDict(
                        [
                            ("plugin", plugin_name),
                            ("event", ""),
                            ("function", ""),
                        ]
                    )
                    table_data.append(row)

            # Create and print table
            table = totable(table_data)
            esg.ui.print(table)


@plugins.command("enable")
@click.argument("plugin_name")
@opts.verbosity
def enable_plugin_cmd(plugin_name: str, verbosity: Verbosity):
    """Enable a plugin."""
    esg = init_esgpull(verbosity=verbosity)
    plugin_path = esg.config.paths.plugins / f"{plugin_name}.py"
    if not plugin_path.exists():
        esg.ui.print(f"Plugin file '{plugin_name}.py' not found.")
        return

    result = esg.plugin_manager.enable_plugin(plugin_name)
    if result:
        esg.ui.print(f"Plugin '{plugin_name}' enabled.")
    else:
        esg.ui.print(f"Failed to enable plugin '{plugin_name}'.")


@plugins.command("disable")
@click.argument("plugin_name")
@opts.verbosity
def disable_plugin_cmd(plugin_name: str, verbosity: Verbosity):
    """Disable a plugin."""
    esg = init_esgpull(verbosity=verbosity)

    # Check if the plugin file exists (even if not loaded)
    plugin_path = esg.config.paths.plugins / f"{plugin_name}.py"
    if not plugin_path.exists():
        esg.ui.print(f"Plugin file '{plugin_name}.py' not found.")
        return

    result = esg.plugin_manager.disable_plugin(plugin_name)
    if result:
        esg.ui.print(f"Plugin '{plugin_name}' disabled.")
    else:
        esg.ui.print(f"Failed to disable plugin '{plugin_name}'.")


@plugins.command("config")
@args.key
@args.value
@opts.default
@opts.generate
@opts.verbosity
def config_plugin(
    verbosity: Verbosity,
    key: str | None,
    value: str | None,
    default: bool,
    generate: bool,
    # config_set: str | None = None,
    # config_unset: str | None = None,
):
    """View or modify plugin configuration."""
    esg = init_esgpull(verbosity=verbosity)

    with esg.ui.logging("plugins", onraise=Abort):
        check_enabled(esg)
        esg.plugin_manager.discover_plugins(
            esg.config.paths.plugins, load_all=True
        )
        if key is not None:
            plugin_name = key.split(".", 1)[0]
            plugin_path = esg.config.paths.plugins / f"{plugin_name}.py"
            if not plugin_path.exists():
                esg.ui.print(f"Plugin file '{plugin_name}.py' not found.")
                raise Abort
        if key is not None and value is not None:
            """update config"""
            if default:
                raise BadOptionUsage(
                    "default",
                    dedent(
                        f"""
                        --default/-d is invalid with a value.
                        Instead use:

                        $ esgpull plugins config {key} -d
                        """
                    ),
                )
            # Split key into plugin_name and remaining_path
            plugin_name, *rest = key.split(".", 1)
            remaining_path = rest[0] if rest else None

            # Set the configuration value
            if remaining_path:
                old_value = esg.plugin_manager.set_plugin_config(
                    plugin_name, remaining_path, value
                )
                info = extract_subdict(esg.plugin_manager.config.plugins, key)
                esg.ui.print(info, toml=True)
                esg.ui.print(f"Previous value: {old_value}")
            else:
                esg.ui.print(f"Cannot set plugin name directly: {plugin_name}")
                raise Abort
        elif key is not None:
            if default:
                # Split key into plugin_name and remaining_path
                plugin_name, *rest = key.split(".", 1)
                remaining_path = rest[0] if rest else None

                if remaining_path:
                    esg.plugin_manager.unset_plugin_config(
                        plugin_name, remaining_path
                    )
                    msg = f":+1: Config reset to default for {key}"
                    esg.ui.print(msg)
                else:
                    esg.ui.print(
                        f"Cannot reset plugin directly: {plugin_name}"
                    )
                    raise Abort
            else:
                config = esg.plugin_manager.config.plugins
                config = extract_subdict(config, key)
                esg.ui.print(config, toml=True)
        elif generate:
            # Write the config file with all current settings
            esg.plugin_manager.write_config(generate_full_config=True)
            msg = f":+1: Plugin config generated at {esg.plugin_manager.config_path}"
            esg.ui.print(msg)

        else:
            esg.ui.rule(str(esg.plugin_manager.config_path))
            esg.ui.print(esg.plugin_manager.config.plugins, toml=True)


@plugins.command("test")
@click.argument(
    "event_type",
    type=click.Choice([e.value for e in Event]),
)
@opts.verbosity
def test_plugin(
    event_type: str,
    verbosity: Verbosity,
):
    """Test plugin events."""
    esg = init_esgpull(verbosity=verbosity)
    with esg.ui.logging("plugins", onraise=Abort):
        check_enabled(esg)
        event = Event(event_type)
        # Create sample test data
        file1 = File(
            file_id="file1",
            dataset_id="dataset",
            master_id="master",
            url="file",
            version="v0",
            filename="file.nc",
            local_path="project/folder",
            data_node="data_node",
            checksum="0",
            checksum_type="0",
            size=2**42,
            status=FileStatus.Queued,
        )
        file2 = file1.clone()
        file2.file_id = "file2"
        file2.status = FileStatus.Done
        error = ValueError("Placeholder example error")
        dataset = Dataset(dataset_id="dataset", total_files=2)
        dataset.files = [file1, file2]
        emit(
            event,
            file=file1,
            dataset=dataset,
            error=error,
        )


@plugins.command("create")
@click.argument(
    "events",
    nargs=-1,
    type=click.Choice([e.value for e in Event]),
)
@click.option("-n", "--name", required=True, type=str)
@opts.verbosity
def create_plugin(
    name: str, verbosity: Verbosity, events: list[str] | None = None
):
    """Create a new plugin template."""
    esg = init_esgpull(verbosity=verbosity)

    # Create plugin file path
    plugin_path = esg.config.paths.plugins / f"{name}.py"
    if plugin_path.exists():
        esg.ui.print(f"Plugin file already exists: {plugin_path}")
        return

    # If no events specified, include them all
    if not events:
        events = [e.value for e in Event]

    # Generate template
    template = f"""# {name} plugin for ESGPull
# 
# This plugin was auto-generated. Edit as needed.
from datetime import datetime
from pathlib import Path
from logging import Logger

from esgpull.models import Dataset, File, Query
from esgpull.plugin import Event, on

# Specify version compatibility (optional)
MIN_ESGPULL_VERSION = "{__version__}"
MAX_ESGPULL_VERSION = None

# Configuration class for the plugin (optional)
class Config:
    \"""Configuration for {name} plugin\"""
    # Add your configuration options here
    timeout = 30  # Example config option
    log_level = "INFO"  # Example config option

"""

    # Add handlers for requested events
    handlers = {
        "file_complete": """
# File complete event handler
@on(Event.file_complete, priority="normal")
def handle_file_complete(
    file: File, 
    destination: Path, 
    start_time: datetime,
    end_time: datetime,
    logger: Logger
):
    \"""Handle file complete event\"""
    logger.info(f"File downloaded: {file.filename}")
    # Add your custom logic here
""",
        "file_error": """
# Download error event handler
@on(Event.file_error, priority="normal")
def handle_file_error(file: File, exception: Exception, logger: Logger):
    \"""Handle file error event\"""
    logger.error(f"Download failed for {file.filename}: {exception}")
    # Add your custom logic here
""",
        "dataset_complete": """
# Dataset complete event handler
@on(Event.dataset_complete, priority="normal")
def handle_dataset_complete(dataset: Dataset, logger: Logger):
    \"""Handle dataset complete event\"""
    logger.error(f"Dataset downloaded: {dataset.dataset_id}")
    # Add your custom logic here
""",
    }

    for event in events:
        template += handlers[event]

    # Write the plugin file
    with open(plugin_path, "w") as f:
        f.write(template)

    esg.ui.print(f"Plugin template created at: {plugin_path}")
    esg.ui.print("Edit the file to implement your custom plugin logic.")



"""update tables

Revision ID: 0.6.2
Revises: 0.6.1
Create Date: 2023-08-18 17:11:15.605131

"""

# revision identifiers, used by Alembic.
revision = "0.6.2"
down_revision = "0.6.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



import pytest

from esgpull.database import Database
from esgpull.graph import Graph
from esgpull.models import Query
from tests.utils import dict_equals_ignore


@pytest.fixture
def base():
    base = Query(
        selection=dict(
            project="CMIP5",
            ensemble="r1i1p1",
            realm="atmos",
        )
    )
    base.compute_sha()
    return base


@pytest.fixture
def a(base):
    a = Query(
        require=base.sha,
        selection=dict(
            experiment=["historical", "rcp26"],
            time_frequency="mon",
            variable="tasmin",
        ),
    )
    a.compute_sha()
    return a


@pytest.fixture
def b(base):
    b = Query(
        require=base.sha,
        selection=dict(
            experiment="rcp85",
            time_frequency="day",
            variable=["tas", "ua"],
        ),
    )
    b.compute_sha()
    return b


@pytest.fixture
def c(base):
    c = Query(
        require=base.sha,
        selection=dict(
            time_frequency=["day", "mon", "fx"],
            variable="tasmax",
        ),
    )
    c.compute_sha()
    return c


@pytest.fixture
def graph(base, a, b, c):
    graph = Graph(db=None)
    graph.add(base, a, b, c)
    return graph


def test_add_all(base, a, b, c):
    graph = Graph(db=None)
    graph.add(base, a, b, c)
    assert len(graph.queries) == 4
    root_queries = graph.get_children(None)
    base_children = graph.get_children(base.sha)
    assert len(graph.queries) == 4
    assert len(root_queries) == 1
    assert len(base_children) == 3


def test_add_1_by_1(base, a, b, c):
    graph = Graph(db=None)
    graph.add(base)
    graph.add(a)
    graph.add(b)
    graph.add(c)
    assert len(graph.queries) == 4
    root_queries = graph.get_children(None)
    base_children = graph.get_children(base.sha)
    assert len(graph.queries) == 4
    assert len(root_queries) == 1
    assert len(base_children) == 3


def test_add_1_by_1_reverse(base, a, b, c):
    graph = Graph(db=None)
    graph.add(c)
    graph.add(b)
    graph.add(a)
    graph.add(base)
    assert len(graph.queries) == 4
    root_queries = graph.get_children(None)
    base_children = graph.get_children(base.sha)
    assert len(graph.queries) == 4
    assert len(root_queries) == 1
    assert len(base_children) == 3


def test_asdict(graph, base, a, b, c):
    dump = graph.asdict()
    expected = {
        base.sha: dict(
            selection=dict(
                project="CMIP5",
                ensemble="r1i1p1",
                realm="atmos",
            )
        ),
        a.sha: dict(
            require=base.sha,
            selection=dict(
                experiment=["historical", "rcp26"],
                time_frequency="mon",
                variable="tasmin",
            ),
        ),
        b.sha: dict(
            require=base.sha,
            selection=dict(
                experiment="rcp85",
                time_frequency="day",
                variable=["tas", "ua"],
            ),
        ),
        c.sha: dict(
            require=base.sha,
            selection=dict(
                time_frequency=sorted(["day", "mon", "fx"]),
                variable="tasmax",
            ),
        ),
    }
    for sha, query_dict in dump.items():
        assert dict_equals_ignore(
            query_dict,
            expected[sha],
            ignore_keys=["added_at", "updated_at"],
        )


def test_dump(graph, base, a, b, c):
    dump = graph.dump()
    assert dict_equals_ignore(
        dump[0],
        dict(
            selection=dict(
                project="CMIP5",
                ensemble="r1i1p1",
                realm="atmos",
            )
        ),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        dump[1],
        dict(
            require=base.sha,
            selection=dict(
                experiment=["historical", "rcp26"],
                time_frequency="mon",
                variable="tasmin",
            ),
        ),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        dump[2],
        dict(
            require=base.sha,
            selection=dict(
                experiment="rcp85",
                time_frequency="day",
                variable=["tas", "ua"],
            ),
        ),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        dump[3],
        dict(
            require=base.sha,
            selection=dict(
                time_frequency=sorted(["day", "mon", "fx"]),
                variable="tasmax",
            ),
        ),
        ignore_keys=["added_at", "updated_at"],
    )


def test_expand(graph, base, a, b, c):
    assert dict_equals_ignore(
        graph.expand(base.sha).asdict(),
        base.asdict(),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        graph.expand(a.sha).asdict(),
        dict(
            selection=dict(
                project="CMIP5",
                ensemble="r1i1p1",
                realm="atmos",
                experiment=["historical", "rcp26"],
                time_frequency="mon",
                variable="tasmin",
            )
        ),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        graph.expand(b.sha).asdict(),
        dict(
            selection=dict(
                project="CMIP5",
                ensemble="r1i1p1",
                realm="atmos",
                experiment="rcp85",
                time_frequency="day",
                variable=["tas", "ua"],
            )
        ),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        graph.expand(c.sha).asdict(),
        dict(
            selection=dict(
                project="CMIP5",
                ensemble="r1i1p1",
                realm="atmos",
                time_frequency=sorted(["day", "mon", "fx"]),
                variable="tasmax",
            )
        ),
        ignore_keys=["added_at", "updated_at"],
    )


@pytest.fixture
def graph2():
    graph = Graph(db=None)
    cordex = Query(selection=dict(project="CORDEX"))
    cordex.compute_sha()
    cmip5 = Query(selection=dict(project="CMIP5"))
    cmip5.compute_sha()
    cmip5_tas = Query(require=cmip5.sha, selection=dict(variable_id="tas"))
    cmip5_tas.compute_sha()
    cmip5_pr = Query(require=cmip5.sha, selection=dict(variable_id="pr"))
    cmip5_pr.compute_sha()
    cmip6 = Query(selection=dict(project="CMIP6"), tags=["cmip6"])
    cmip6.compute_sha()
    cmip6.sha = "bad_value"
    cmip6_tas = Query(
        require=cmip6.sha, selection=dict(variable_id="tas"), tags=["children"]
    )
    cmip6_tas.compute_sha()
    cmip6_pr = Query(
        require=cmip6.sha, selection=dict(variable_id="pr"), tags=["children"]
    )
    cmip6_pr.compute_sha()
    cmip6_pr_member = Query(
        require=cmip6_pr.sha, selection=dict(member_id="r1i1p1f1")
    )
    cmip6_pr_member.compute_sha()
    graph.add(
        cordex,
        cmip5,
        cmip5_tas,
        cmip5_pr,
        cmip6,
        cmip6_tas,
        cmip6_pr,
        cmip6_pr_member,
        clone=False,
    )
    return graph


def test_children(graph2):
    assert len(graph2.get_children("bad_value")) == 2
    assert len(graph2.get_all_children("bad_value")) == 3


@pytest.fixture
def db(config):
    return Database.from_config(config)


def test_children_db(graph2, db):
    graph = Graph(db)
    graph.add(*graph2.queries.values(), clone=False)
    graph.merge()
    graph = Graph(db)  # reset graph
    assert len(graph.queries) == 0
    assert len(graph.get_children("bad_value")) == 2
    assert len(graph.get_all_children("bad_value")) == 3


def test_delete_add_merge(base, db):
    graph1 = Graph(db)
    graph1.add(base)
    graph1.merge()
    graph2 = Graph(db)
    assert base in graph2
    graph2.delete(base)
    graph2.add(base)
    graph2.merge()
    graph3 = Graph(db)
    assert base in graph3



from dataclasses import dataclass, field
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class Result(Generic[T]):
    data: T
    ok: bool
    err: BaseException | None


@dataclass
class Ok(Result[T]):
    ok: bool = field(default=True, init=False)
    err: None = field(default=None, init=False)


@dataclass
class Err(Result[T]):
    ok: bool = field(default=False, init=False)
    err: BaseException = field()


# class Result:
#     ok: bool
#     file: File
#     completed: int
#     err: BaseException | None


# class Ok(Result):
#     ok = True
#     file: File
#     completed: int
#     err = None

# def __init__(self, file: File, completed: int) -> None:
#     self.file = file
#     self.completed = completed


# class Err(Result):
#     ok = False
#     file: File
#     completed: int
#     err: BaseException

#     def __init__(self, file: File, completed: int, err: BaseException) -> None:
#         self.file = file
#         self.completed = completed
#         self.err = err



"""update tables

Revision ID: 0.7.0
Revises: 0.6.5
Create Date: 2024-09-12 16:32:48.745570

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0.7.0'
down_revision = '0.6.5'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.alter_column('size',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.alter_column('size',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=False)

    # ### end Alembic commands ###



from pathlib import Path

from click.testing import CliRunner

from esgpull import Esgpull
from esgpull.cli.download import download
from esgpull.config import Config
from esgpull.models import File, FileStatus, Query


def test_download_errors_in_progress(root: Path, config: Config):
    # Configure the environment
    config.generate(overwrite=True)
    runner = CliRunner()
    esg = Esgpull(root)

    # Create a query without updating (no datasets will be created)
    query = Query()
    query.compute_sha()
    query.track(query.options.default())

    # Manually create files
    failing_file = File(
        file_id="failing_file",
        dataset_id="test_dataset_id",
        master_id="test_master_id",
        url="https://nonexistent.path/to/file.nc",
        version="1.0",
        filename="failing.nc",
        local_path="test/failing",
        data_node="test_node",
        checksum="12345",
        checksum_type="SHA256",
        size=1000,
        status=FileStatus.Queued,
    )
    failing_file.compute_sha()

    # Add files to query
    query.files.append(failing_file)

    # Add query to database
    esg.db.add(query)

    # Run the download instruction in the CLI
    result_download = runner.invoke(download)

    # Parse the output to extract
    output = result_download.output
    assert "download failed" in output



# Plugin with incompatible version requirements for testing

from esgpull.plugin import Event, on

# Version compatibility - requiring a future version
MIN_ESGPULL_VERSION = (
    "999.0.0"  # Intentionally using a version that doesn't exist
)
MAX_ESGPULL_VERSION = None


# Configuration class
class Config:
    """Configuration for incompatible_plugin"""

    log_level = "INFO"


# Track calls for testing
calls = {"file_complete": 0}


@on(Event.file_complete)
def handle_file_complete(file, destination, start_time, end_time, logger):
    """Handle file complete event"""
    logger.info(f"This shouldn't run: {file.filename}")
    calls["file_complete"] += 1



from contextlib import nullcontext as does_not_raise
from time import perf_counter

import pytest

from esgpull.context import Context, IndexNode, _distribute_hits_impl
from esgpull.models import Query
from tests.utils import CEDA_NODE, DRKZ_NODE, IPSL_NODE, ORNL_BRIDGE


@pytest.fixture
def ctx(config):
    return Context(config=config)


@pytest.fixture
def empty():
    return Query()


@pytest.fixture
def cmip6_ipsl():
    query = Query()
    query.options.distrib = False
    query.selection.mip_era = "CMIP6"
    query.selection.institution_id = "IPSL"
    return query


@pytest.fixture(params=["empty", "cmip6_ipsl"])
def query(request):
    return request.getfixturevalue(request.param)


parametrized_index = pytest.mark.parametrize("index", [CEDA_NODE, ORNL_BRIDGE])


def test_multi_index(ctx, empty):
    index_nodes = [CEDA_NODE, DRKZ_NODE]
    results = []
    for index_node in index_nodes:
        query_results = ctx.prepare_hits(
            empty,
            file=False,
            index_node=index_node,
        )
        results.extend(query_results)
    assert len(results) == 2
    for result, index_node in zip(results, index_nodes):
        assert index_node in str(result.request.url)
        assert index_node == result.request.headers["host"]


def test_adjust_hits(ctx):
    variable_ids = ["tas", "tasmin"]
    queries = []
    for variable_id in variable_ids:
        query = Query(selection=dict(variable_id=variable_id))
        queries.append(query)
    hits = ctx.hits(*queries, file=False)
    page_limit = 5
    first_20 = ctx.prepare_search(
        *queries,
        file=False,
        hits=hits,
        page_limit=page_limit,
        max_hits=len(variable_ids) * page_limit * 2,
    )
    assert len(first_20) >= len(variable_ids) * 2
    variable_offsets = {variable_id: 0 for variable_id in variable_ids}
    for result in first_20:
        variable_id = result.query.selection.variable_id[0]
        params = dict(result.request.url.params.items())

        assert int(params["offset"]) == variable_offsets[variable_id]
        assert int(params["limit"]) <= page_limit
        assert params["query"] == f"variable_id:{variable_id}"

        variable_offsets[variable_id] += page_limit

    offset_100 = ctx.prepare_search(
        *queries,
        file=False,
        hits=hits,
        offset=100,
        page_limit=page_limit,
        max_hits=len(variable_ids) * page_limit * 2,
    )
    assert len(offset_100) >= len(variable_ids) * 2

    # ensure offsets follow the proportional splitter: sum matches the limit and each bucket stays within 1 of its ideal share
    ideal = {
        variable_id: hits[i] / sum(hits) * 100
        for i, variable_id in enumerate(variable_ids)
    }
    distributed = _distribute_hits_impl(hits, 100)
    variable_offsets = {
        variable_id: distributed[i]
        for i, variable_id in enumerate(variable_ids)
    }
    for result in offset_100:
        variable_id = result.query.selection.variable_id[0]
        params = dict(result.request.url.params.items())
        offset = int(params["offset"])

        assert offset == variable_offsets[variable_id]
        # invariant: within 1 of ideal
        assert abs(offset - ideal[variable_id]) < 1
        assert int(params["limit"]) <= page_limit
        assert params["query"] == f"variable_id:{variable_id}"

        variable_offsets[variable_id] += page_limit
        ideal[variable_id] += page_limit


class Timer:
    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, *excs):
        self.duration = perf_counter() - self.start


# @pytest.mark.slow
# @pytest.mark.xfail(
#     raises=ValueError,
#     reason="ESGF bridge API gives index_node values that are not valid URLs",
# )
# def test_search_distributed(ctx):
#     query = Query()
#     # ctx.config.api.http_timeout = 60
#     query.options.distrib = True
#     # configure terms to target ~1000 files across different index nodes
#     query.selection.mip_era = "CMIP6"
#     query.selection.table_id = "Amon"
#     query.selection.variable_id = "tas"
#     query.selection.member_id = "r20i1p1f1"
#     with Timer() as t_regular:
#         datasets_regular = ctx.datasets(
#             query,
#             max_hits=None,
#             keep_duplicates=True,
#         )
#     with Timer() as t_distributed:
#         hints = ctx.hints(query, file=False, facets=["index_node"])
#         results = ctx.prepare_search_distributed(
#             query,
#             file=False,
#             hints=hints,
#             max_hits=None,
#         )
#         coro = ctx._datasets(*results, keep_duplicates=True)
#         datasets_distributed = ctx._sync(coro)
#     dataset_ids_regular = {d.dataset_id for d in datasets_regular}
#     dataset_ids_distributed = {d.dataset_id for d in datasets_distributed}
#     assert dataset_ids_regular == dataset_ids_distributed
#     # assert t_regular.duration >= t_distributed.duration
#     logging.info(f"{t_regular.duration}")
#     logging.info(f"{t_distributed.duration}")


@parametrized_index
def test_ipsl_hits_exist(ctx, index: str, cmip6_ipsl):
    hits = ctx.hits(
        cmip6_ipsl,
        file=False,
        index_node=index,
    )
    assert 1_000 < hits[0]


@parametrized_index
def test_more_files_than_datasets(ctx, index: str, query):
    assert sum(ctx.hits(query, file=False)) < sum(ctx.hits(query, file=True))


@parametrized_index
@pytest.mark.slow
def test_hints(ctx, index: str, cmip6_ipsl):
    facets = ["institution_id", "variable_id"]
    hints = ctx.hints(cmip6_ipsl, file=False, facets=facets)[0]
    assert list(hints["institution_id"]) == cmip6_ipsl.selection.institution_id
    assert len(hints["variable_id"]) > 1


def test_hits_from_hints(ctx):
    hints = {"facet_name": {"value_a": 1, "value_b": 2, "value_c": 3}}
    hits = ctx.hits_from_hints(hints)
    assert hits == [6]


@parametrized_index
@pytest.mark.parametrize(
    "query_all",
    [
        Query(),
        Query(selection={"variable_id": "tas"}),
        Query(selection={"experiment_id": "ssp*", "variable_id": "tas"}),
    ],
)
def test_ignore_facet_hits(ctx, index: str, query_all: Query):
    query_ipsl = Query(selection={"institution_id": "IPSL"}) << query_all
    query_not_ipsl = Query(selection={"!institution_id": "IPSL"}) << query_all
    hits_all = ctx.hits(query_all, file=False, index_node=index)[0]
    hits_ipsl = ctx.hits(query_ipsl, file=False, index_node=index)[0]
    hits_not_ipsl = ctx.hits(query_not_ipsl, file=False, index_node=index)[0]
    assert all(hits > 0 for hits in [hits_all, hits_ipsl, hits_not_ipsl])
    assert hits_all == hits_ipsl + hits_not_ipsl


@pytest.mark.parametrize(
    "index,url,is_bridge",
    [
        (
            IPSL_NODE,
            f"https://{IPSL_NODE}/esg-search/search",
            False,
        ),
        (
            CEDA_NODE,
            f"https://{CEDA_NODE}/esg-search/search",
            False,
        ),
        (
            f"https://{IPSL_NODE}/esg-search/search",
            f"https://{IPSL_NODE}/esg-search/search",
            False,
        ),
        (
            f"https://{CEDA_NODE}/esg-search/search",
            f"https://{CEDA_NODE}/esg-search/search",
            False,
        ),
        (
            ORNL_BRIDGE,
            f"https://{ORNL_BRIDGE}",
            True,
        ),
        (
            f"https://{ORNL_BRIDGE}",
            f"https://{ORNL_BRIDGE}",
            True,
        ),
    ],
)
def test_index2url(index: str, url: str, is_bridge: bool):
    for value in (index, url):
        index_node = IndexNode(value=value)
        assert index_node.url == url
        assert index_node.is_bridge() == is_bridge


@pytest.mark.parametrize(
    "queries",
    [
        [],
        [Query()],
        [
            Query(
                selection=dict(
                    project="CMIP6",
                    institution_id="IPSL",
                    variable_id="uv",
                )
            )
        ],
        [
            Query(),
            Query(
                selection=dict(
                    project="CMIP6",
                    institution_id="IPSL",
                    variable_id="uv",
                )
            ),
        ],
        [Query(selection=dict(project="notaproject"))],
    ],
)
@pytest.mark.parametrize("file", [True, False])
@pytest.mark.parametrize(
    "index_node",
    ## TODO: test bridge, but it is super slow
    [
        IPSL_NODE,
        CEDA_NODE,
        "https://github.com",
        "not_a_real.url",
    ],
)
def test_hits_never_empty(
    ctx: Context,
    queries: tuple[Query],
    file: bool,
    index_node: str,
):
    ctx.noraise = True
    hits = ctx.hits(*queries, file=file, index_node=index_node)
    assert len(hits) == len(queries)


@pytest.mark.parametrize(
    ("index_node", "exc"),
    ## TODO: test bridge, but it is super slow
    [
        pytest.param(
            IPSL_NODE,
            does_not_raise(),
            marks=pytest.mark.xfail(reason="unstable"),
        ),
        (CEDA_NODE, does_not_raise()),
        ("https://github.com", pytest.raises(Exception)),
        ("not_a_real.url", pytest.raises(Exception)),
    ],
)
def test_probe(
    ctx: Context,
    index_node: str,
    exc,
):
    ctx.config.api.index_node = index_node
    with exc:
        ctx.probe()


def test_bridge_exact_match_params(ctx):
    query = Query(selection=dict(source_id="CESM2", variable_id="tas"))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=ORNL_BRIDGE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "source_id" in params
    assert params["source_id"] == "CESM2"
    assert "variable_id" in params
    assert params["variable_id"] == "tas"
    assert "query" not in params or params["query"] == ""


def test_bridge_wildcard_query_param(ctx):
    query = Query(selection=dict(source_id="CESM*", variable_id="tas*"))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=ORNL_BRIDGE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "query" in params
    assert "source_id" not in params
    assert "variable_id" not in params
    assert "source_id:CESM*" in params["query"]
    assert "variable_id:tas*" in params["query"]


def test_bridge_mixed_exact_wildcard(ctx):
    query = Query(selection=dict(source_id="CESM2", variable_id="tas*"))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=ORNL_BRIDGE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "source_id" in params
    assert params["source_id"] == "CESM2"
    assert "query" in params
    assert "variable_id:tas*" in params["query"]
    assert "variable_id" not in params


def test_bridge_multi_value_exact(ctx):
    query = Query(selection=dict(source_id=["CESM2", "CESM2-LENS2"]))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=ORNL_BRIDGE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "source_id" in params
    assert params["source_id"] == "CESM2,CESM2-LENS2"
    assert "query" not in params or params["query"] == ""


def test_bridge_negated_query(ctx):
    query = Query(selection=dict(**{"!institution_id": "IPSL"}))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=ORNL_BRIDGE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "query" in params
    assert "institution_id" not in params
    assert 'NOT (institution_id:"IPSL")' in params["query"]


def test_bridge_mixed_wildcard_warning(ctx, caplog):
    query = Query(selection=dict(source_id=["CESM2", "CESM*"]))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=ORNL_BRIDGE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "source_id" not in params
    assert "query" in params
    assert "source_id:" in params["query"]
    assert any(
        "source_id has mixed wildcard/non-wildcard values" in record.message
        for record in caplog.records
    )


def test_solr_unchanged(ctx):
    query = Query(selection=dict(source_id="CESM2", variable_id="tas"))
    result = ctx.prepare_hits(
        query,
        file=False,
        index_node=IPSL_NODE,
    )[0]
    params = dict(result.request.url.params.items())

    assert "query" in params
    assert params["query"] == "source_id:CESM2 AND variable_id:tas"
    assert "source_id" not in params
    assert "variable_id" not in params


def test_files_skips_duplicate_file_ids(ctx):
    """Test that _files skips duplicate file_ids to prevent DB constraint violations."""
    import asyncio
    from unittest.mock import patch, MagicMock
    from esgpull.context import ResultFiles

    # Create Solr response docs that will be deserialized by File.serialize()
    # Two files with same file_id (dataset.v1.file.nc) but different checksums
    solr_docs = [
        {
            "instance_id": "dataset.v1.file.nc|node1.com",
            "dataset_id": "dataset.v1|node1.com",
            "title": "file.nc",
            "url": "https://node1.com/file.nc|application/netcdf",
            "data_node": "node1.com",
            "checksum": "abc123",
            "checksum_type": "SHA256",
            "size": 1000,
            "directory_format_template_": "%(root)s/v1",  # Simple template
        },
        {
            "instance_id": "dataset.v1.file.nc|node2.com",
            "dataset_id": "dataset.v1|node2.com",
            "title": "file.nc",
            "url": "https://node2.com/file.nc|application/netcdf",
            "data_node": "node2.com",
            "checksum": "def456",  # Different checksum!
            "checksum_type": "SHA256",
            "size": 1000,
            "directory_format_template_": "%(root)s/v1",  # Simple template
        },
        {
            "instance_id": "dataset.v1.other.nc|node1.com",
            "dataset_id": "dataset.v1|node1.com",
            "title": "other.nc",
            "url": "https://node1.com/other.nc|application/netcdf",
            "data_node": "node1.com",
            "checksum": "xyz789",
            "checksum_type": "SHA256",
            "size": 2000,
            "directory_format_template_": "%(root)s/v1",  # Simple template
        },
    ]

    # Create a ResultFiles with proper JSON that will be parsed by process()
    query = Query()
    result = ResultFiles(query=query, file=True)
    result.request = MagicMock()
    result.json = {"response": {"docs": solr_docs}}

    # Mock _fetch to yield our result
    async def mock_fetch(*results):
        yield result

    async def run_test():
        with patch.object(ctx, "_fetch", mock_fetch):
            return await ctx._files(result, keep_duplicates=False)

    result_files = asyncio.run(run_test())

    # Should only have 2 files (second one with duplicate file_id is skipped)
    assert len(result_files) == 2

    # Verify the file_ids in the result
    file_ids = {f.file_id for f in result_files}
    assert file_ids == {"dataset.v1.file.nc", "dataset.v1.other.nc"}

    # The first file with dataset.v1.file.nc should be kept (abc123 checksum)
    duplicate_files = [
        f for f in result_files if f.file_id == "dataset.v1.file.nc"
    ]
    assert len(duplicate_files) == 1
    assert duplicate_files[0].checksum == "abc123"



"""update tables

Revision ID: 0.8.0
Revises: 14c72daea083
Create Date: 2025-05-15 11:28:10.755003

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0.8.0'
down_revision = '14c72daea083'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



import sys
from collections import OrderedDict
from collections.abc import MutableMapping, Sequence
from enum import Enum
from pathlib import Path
from typing import Any, Literal

import click
import yaml
from click.exceptions import BadArgumentUsage
from rich.box import MINIMAL_DOUBLE_HEAD
from rich.table import Table
from rich.text import Text

from esgpull import Esgpull
from esgpull.graph import Graph
from esgpull.models import Dataset, File, Option, Options, Query, Selection
from esgpull.tui import UI, TempUI, Verbosity, logger
from esgpull.utils import format_size


def get_command() -> Text:
    exe, *args = sys.argv
    args = [arg for arg in args if arg != "--record"]
    return Text(" ".join(["$", Path(exe).name, *args]))


def init_esgpull(
    verbosity: Verbosity,
    safe: bool = True,
    record: bool = False,
    load_db: bool = True,
    no_default_query: bool = False,
) -> Esgpull:
    TempUI.verbosity = Verbosity.Errors
    with TempUI.logging():
        esg = Esgpull(
            verbosity=verbosity,
            safe=safe,
            record=record,
            load_db=load_db,
        )
        if no_default_query:
            esg.config.api.default_query_id = ""
        if record:
            esg.ui.print(get_command())
    return esg


class Messages:
    @staticmethod
    def no_such_query(name: str) -> str:
        return f":stop_sign: No such query [green]{name}[/]"

    @staticmethod
    def none_tagged(tag: str) -> str:
        return f":stop_sign: No query tagged with [magenta]{tag}[/]"

    @staticmethod
    def multimatch(name: str) -> str:
        return f"Found multiple queries starting with [b cyan]{name}[/]"


class EnumParam(click.Choice):
    name = "enum"

    def __init__(self, enum: type[Enum]):
        self.__enum = enum
        super().__init__(choices=[item.value for item in enum])

    def convert(self, value, param, ctx) -> Enum:
        converted_str = super().convert(value, param, ctx)
        return self.__enum(converted_str)


def filter_keys(
    docs: Sequence[File | Dataset],
    ids: range,
    size: bool = True,
    data_node: bool = False,
    # date: bool = False,
) -> list[OrderedDict[str, Any]]:
    result: list[OrderedDict[str, Any]] = []
    for i, doc in zip(ids, docs):
        od: OrderedDict[str, Any] = OrderedDict()
        od["id"] = str(i)
        if isinstance(doc, File):
            od["file"] = doc.file_id
        else:
            od["dataset"] = doc.dataset_id
            od["#"] = str(doc.number_of_files)
        if size:
            od["size"] = doc.size
        if data_node:
            od["data_node"] = doc.data_node
        # if date:
        #     od["date"] = doc.get("timestamp") or doc.get("_timestamp")
        result.append(od)
    return result


def totable(docs: list[OrderedDict[str, Any]]) -> Table:
    table = Table(box=MINIMAL_DOUBLE_HEAD, show_edge=False)
    for key in docs[0].keys():
        justify: Literal["left", "right", "center"]
        if key in ["file", "dataset", "plugin"]:
            justify = "left"
        else:
            justify = "right"
        table.add_column(
            Text(key, justify="center"),
            justify=justify,
        )
    for doc in docs:
        row: list[str] = []
        for key, value in doc.items():
            if key == "size":
                value = format_size(value)
            row.append(value)
        table.add_row(*row)
    return table


def safe_value(value: str) -> str:
    if " " in value:
        return f'"{value}"'
    else:
        return value


def parse_facets(facets: list[str]) -> Selection:
    facet_dict: dict[str, list[str]] = {}
    for facet in facets:
        match facet.split(":"):
            case [value]:
                name = "query"
            case [name, value] if name and value:
                ...
            case _:
                raise BadArgumentUsage(f"{facet!r} is not valid syntax.")
        values = list(map(safe_value, value.split(",")))
        facet_dict.setdefault(name, [])
        facet_dict[name].extend(values)
    selection = Selection()
    for name, values in facet_dict.items():
        selection[name] = values
    return selection


def parse_query(
    facets: list[str],
    # query options
    tags: list[str],
    require: str | None,
    distrib: str | None,
    latest: str | None,
    replica: str | None,
    retracted: str | None,
) -> Query:
    logger.info(f"{facets=}")
    logger.info(f"{tags=}")
    logger.info(f"{require=}")
    logger.info(f"{distrib=}")
    logger.info(f"{latest=}")
    logger.info(f"{replica=}")
    logger.info(f"{retracted=}")
    options = Options(
        distrib=distrib or Option.notset,
        latest=latest or Option.notset,
        replica=replica or Option.notset,
        retracted=retracted or Option.notset,
    )
    selection = parse_facets(facets)
    return Query(
        tags=tags,
        require=require,
        options=options,
        selection=selection,
    )


def is_list_of_maps(seq: Sequence) -> bool:
    return all(isinstance(item, MutableMapping) for item in seq)


def serialize_queries_from_file(path: Path) -> list[Query]:
    with path.open() as f:
        content = yaml.safe_load(f)
    queries: list[MutableMapping]
    if isinstance(content, list):
        queries = content
    elif isinstance(content, MutableMapping):
        values = list(content.values())
        if is_list_of_maps(values):
            queries = values
        else:
            queries = [content]
    else:
        raise ValueError(content)
    return [Query(**query) for query in queries]


def valid_name_tag(
    graph: Graph,
    ui: UI,
    query_id: str | None,
    tag: str | None,
) -> bool:
    result = True
    if query_id is not None:
        shas = graph.matching_shas(query_id, graph._shas)
        if len(shas) > 1:
            ui.print(Messages.multimatch(query_id))
            ui.print(shas, json=True)
            result = False
        elif len(shas) == 0:
            ui.print(Messages.no_such_query(query_id), err=True)
            result = False
    elif tag is not None:
        tags = [t.name for t in graph.get_tags()]
        if tag not in tags:
            ui.print(Messages.none_tagged(tag), err=True)
            result = False
    return result


def get_queries(
    graph: Graph,
    query_id: str | None,
    tag: str | None,
    children: bool = False,
) -> list[Query]:
    queries: list[Query] = []
    if query_id is not None:
        query = graph.get(name=query_id)
        if query is not None:
            queries = [query]
    elif tag is not None:
        queries = graph.with_tag(tag)
    if children:
        for i in range(len(queries)):
            query = queries[0]
            kids = graph.get_all_children(query.sha)
            queries.extend(kids)
    return queries


def extract_subdict(doc: dict, key: str | None) -> dict:
    if key is None:
        return doc
    for part in key.split("."):
        if not part:
            raise KeyError(key)
        elif part in doc:
            doc = doc[part]
        else:
            raise KeyError(part)
    for part in key.split(".")[::-1]:
        doc = {part: doc}
    return doc



from collections.abc import Callable
from pathlib import Path
from typing import Any, TypeAlias, TypeVar

import click
from click_params import StringListParamType

from esgpull.cli.utils import EnumParam
from esgpull.models import FileStatus, Option
from esgpull.tui import Verbosity

F = TypeVar("F", bound=Callable[..., Any])
Dec: TypeAlias = Callable[[F], F]


def compose(*decs: Dec) -> Dec:
    def deco(fn: Callable):
        for dec in reversed(decs):
            fn = dec(fn)
        return fn

    return deco


class args:
    facets: Dec = click.argument(
        "facets",
        type=str,
        nargs=-1,
    )
    key: Dec = click.argument(
        "key",
        type=str,
        nargs=1,
        required=False,
        default=None,
    )
    value: Dec = click.argument(
        "value",
        type=str,
        nargs=1,
        required=False,
        default=None,
    )
    path: Dec = click.argument(
        "path",
        type=click.Path(exists=False, path_type=Path),
        required=False,
        default=None,
    )
    status: Dec = click.argument(
        "status",
        type=EnumParam(FileStatus),
        nargs=-1,
    )
    query_id: Dec = click.argument(
        "query_id",
        type=str,
        nargs=1,
        required=False,
    )
    query_id_required: Dec = click.argument(
        "query_id",
        type=str,
        nargs=1,
        required=True,
    )
    query_ids: Dec = click.argument(
        "query_ids",
        type=str,
        nargs=-1,
        required=True,
    )


class opts:
    all: Dec = click.option(
        "all_",
        "--all",
        "-a",
        is_flag=True,
        default=False,
    )
    children: Dec = click.option(
        "--children",
        "-c",
        is_flag=True,
        default=False,
    )
    default: Dec = click.option(
        "--default",
        "-d",
        is_flag=True,
        default=False,
    )
    detail: Dec = click.option(
        "--detail",
        type=int,
        default=None,
    )
    disable_ssl: Dec = click.option(
        "--disable-ssl",
        is_flag=True,
        default=False,
    )
    dry_run: Dec = click.option(
        "--dry-run",
        "-z",
        is_flag=True,
        default=False,
    )
    facets_hints: Dec = click.option(
        "facets_hints",
        "--facets",
        "-F",
        is_flag=True,
        default=False,
    )
    file: Dec = click.option(
        "--file",
        "-f",
        is_flag=True,
        default=False,
    )
    files: Dec = click.option(
        "--files",
        is_flag=True,
        default=False,
    )
    force: Dec = click.option(
        "--force",
        is_flag=True,
        default=False,
    )
    generate: Dec = click.option(
        "--generate",
        is_flag=True,
        default=False,
    )
    hints: Dec = click.option(
        "--hints",
        "-H",
        type=StringListParamType(","),
        default=None,
    )
    name: Dec = click.option(
        "--name",
        "-n",
        type=str,
        default=None,
    )
    no_default_query: Dec = click.option(
        "--no-default-query",
        is_flag=True,
        default=False,
    )
    query_file: Dec = click.option(
        "--query-file",
        "-q",
        type=click.Path(exists=True, dir_okay=False, path_type=Path),
        default=None,
    )
    quiet: Dec = click.option(
        "--quiet",
        is_flag=True,
        default=False,
    )
    record: Dec = click.option(
        "--record",
        is_flag=True,
        default=False,
    )
    reset: Dec = click.option(
        "--reset",
        is_flag=True,
        default=False,
    )
    shas: Dec = click.option(
        "--shas",
        "-s",
        is_flag=True,
        default=False,
    )
    show: Dec = click.option(
        "--show",
        is_flag=True,
        default=False,
    )
    simple: Dec = click.option(
        "--simple",
        is_flag=True,
        default=False,
    )
    since: Dec = click.option(
        "--since",
        type=str,
        default=None,
    )
    status: Dec = click.option(
        "--status",
        type=EnumParam(FileStatus),
        default=None,
        multiple=True,
    )
    tag: Dec = click.option(
        "--tag",
        "-t",
        type=str,
        default=None,
    )
    track: Dec = click.option(
        "--track",
        is_flag=True,
        default=False,
    )
    yes: Dec = click.option(
        "--yes",
        "-y",
        is_flag=True,
        default=False,
    )
    verbosity: Dec = click.option(
        "verbosity",
        "-v",
        count=True,
        type=EnumParam(Verbosity),
    )


# Display group
_page: Dec = click.option(
    "--page",
    "-p",
    type=int,
    default=0,  # 1?
)
_zero: Dec = click.option(
    "--zero",
    "-0",
    is_flag=True,
    default=False,
)
_all: Dec = click.option(
    "_all",
    "--all",
    "-a",
    is_flag=True,
    default=False,
)

# Json/Yaml exclusive group
_json: Dec = click.option(
    "--json",
    is_flag=True,
    default=False,
)
_yaml: Dec = click.option(
    "--yaml",
    is_flag=True,
    default=False,
)

# Query definition group
OptionChoice = click.Choice([opt.name for opt in list(Option)[:-1]])
_tag: Dec = click.option("tags", "--tag", "-t", multiple=True)
_require: Dec = click.option(
    "--require",
    "-r",
    type=str,
    nargs=1,
    required=False,
    default=None,
)
_distrib: Dec = click.option(
    "--distrib",
    "-d",
    type=OptionChoice,
)
_latest: Dec = click.option(
    "--latest",
    type=OptionChoice,
)
_replica: Dec = click.option(
    "--replica",
    type=OptionChoice,
)
_retracted: Dec = click.option(
    "--retracted",
    type=OptionChoice,
)

# Query dates group
datetime_type = click.DateTime(["%Y-%m-%d"])
_from: Dec = click.option(
    "date_from",
    "--from",
    type=datetime_type,
    default=None,
)
_to: Dec = click.option(
    "date_to",
    "--to",
    type=datetime_type,
    default=None,
)

_children: Dec = click.option(
    "--children",
    "-c",
    is_flag=True,
    default=False,
)
_parents: Dec = click.option(
    "--parents",
    "-p",
    is_flag=True,
    default=False,
)
_expand: Dec = click.option(
    "--expand",
    "-e",
    is_flag=True,
    default=False,
)


class groups:
    display: Dec = compose(
        _page,
        _zero,
        _all,
    )
    json_yaml: Dec = compose(
        _json,
        _yaml,
    )
    query_def: Dec = compose(
        _tag,
        _require,
        _distrib,
        _latest,
        _replica,
        _retracted,
    )
    query_date: Dec = compose(
        _from,
        _to,
    )
    show: Dec = compose(
        _children,
        _parents,
        _expand,
    )



import pytest

from esgpull.exceptions import AlreadySetFacet
from esgpull.models import Query, Tag
from tests.utils import dict_equals_ignore


def test_empty_asdict():
    assert dict_equals_ignore(
        Query().asdict(), {}, ignore_keys=["added_at", "updated_at"]
    )


def test_clone_is_deepcopy():
    query = Query(selection=dict(project="CMIP6"))
    clone = query.clone()
    assert dict_equals_ignore(
        query.asdict(),
        clone.asdict(),
        ignore_keys=["added_at", "updated_at"],
    )
    clone.selection.variable_id = "tas"
    assert not dict_equals_ignore(
        query.asdict(),
        clone.asdict(),
        ignore_keys=["added_at", "updated_at"],
    )


def test_combine():
    # various ways to create subqueries
    a = Query(selection=dict(project="CMIP6"))
    b = Query(selection=dict(mip_era="CMIP6"))
    c = Query(options=dict(distrib=None))
    d = Query(options=dict(distrib=True))
    ab_dict = (a << b).selection.asdict()
    ba_dict = (b << a).selection.asdict()
    abcd_dict = (a << b << c << d).asdict()
    dcba_dict = (d << c << b << a).asdict()
    assert dict_equals_ignore(
        ab_dict, ba_dict, ignore_keys=["added_at", "updated_at"]
    )
    assert dict_equals_ignore(
        ab_dict,
        dict(project="CMIP6", mip_era="CMIP6"),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        abcd_dict,
        dict(selection=ab_dict, options=dict(distrib=True)),
        ignore_keys=["added_at", "updated_at"],
    )
    assert dict_equals_ignore(
        dcba_dict,
        dict(selection=ab_dict, options=dict(distrib=None)),
        ignore_keys=["added_at", "updated_at"],
    )


def test_combine_raise():
    a = Query(selection=dict(project="CMIP5"))
    b = Query(selection=dict(project="CMIP6"))
    with pytest.raises(AlreadySetFacet):
        a << b


def test_combine_removes_require():
    a = Query(selection=dict(project="CMIP6"))
    a.compute_sha()
    b = Query(require=a.sha, selection=dict(variable_id="tas"))
    ab = a << b
    ba = b << a
    assert ab.require is None
    assert ba.require == a.sha


def test_set_tags_raise():
    query = Query()
    with pytest.raises(TypeError):
        query.tags = "tag"
    with pytest.raises(TypeError):
        query.tags = Tag(name="tag")
    with pytest.raises(AttributeError):
        query.tags = ["tag"]
        query.compute_sha()


def test_set_tags_ok():
    query = Query()
    query.tags.append(Tag(name="tag"))
    query_dict = query.asdict()
    assert dict_equals_ignore(
        query_dict, {"tags": "tag"}, ignore_keys=["added_at", "updated_at"]
    )
    query_copy = Query(**query_dict)
    query.compute_sha()
    query_copy.compute_sha()
    assert query.tags[0].sha == query_copy.tags[0].sha



from __future__ import annotations

from collections.abc import Iterator, MutableMapping
from typing import ClassVar, TypeAlias

import sqlalchemy as sa
from rich.pretty import pretty_repr
from sqlalchemy.orm import Mapped, relationship

from esgpull.exceptions import AlreadySetFacet, DuplicateFacet
from esgpull.models.base import Base, Sha
from esgpull.models.facet import Facet

FacetValues: TypeAlias = str | list[str]

selection_facet_proxy = sa.Table(
    "selection_facet",
    Base.metadata,
    sa.Column(
        "selection_sha", Sha, sa.ForeignKey("selection.sha"), primary_key=True
    ),
    sa.Column("facet_sha", Sha, sa.ForeignKey("facet.sha"), primary_key=True),
)


def opposite(facet_name: str) -> str:
    if facet_name[0] == "!":
        return facet_name[1:]
    else:
        return f"!{facet_name}"


class Selection(Base):
    __tablename__ = "selection"
    _facet_names: ClassVar[set[str]] = set()

    _facets: Mapped[list[Facet]] = relationship(
        secondary=selection_facet_proxy,
        default_factory=list,
    )

    @classmethod
    def _add_property(cls, name: str) -> None:
        def getter(self: Selection) -> list[str]:
            indices = self._facet_map_.get(name, set())
            return sorted([self._facets[idx].value for idx in indices])

        def setter(self: Selection, values: FacetValues):
            other = opposite(name)
            if name in self._facet_map_:
                raise AlreadySetFacet(name, ", ".join(self[name]))
            elif other in self._facet_map_:
                raise AlreadySetFacet(other, ", ".join(self[other]))
            facet_map_name = set()
            if isinstance(values, str):
                iter_values = enumerate([values])
            else:
                iter_values = enumerate(values)
            offset = len(self._facets)
            for i, value in iter_values:
                facet = Facet(name=name, value=value)
                if facet in self._facets:
                    raise DuplicateFacet(
                        facet.name,
                        facet.value,
                        pretty_repr(self),
                    )
                self._facets.append(facet)
                facet_map_name.add(offset + i)
            self._facet_map_[name] = facet_map_name

        setattr(cls, name, property(getter, setter))

    @classmethod
    def reset(cls) -> None:
        cls.configure(*DefaultFacets, *BaseFacets, replace=True)

    @classmethod
    def configure(cls, *names: str, replace: bool = True) -> None:
        nameset = set(names) | {f"!{name}" for name in names}
        if replace:
            for name in cls._facet_names:
                delattr(cls, name)
            new_names = nameset
            cls._facet_names = nameset
        else:
            new_names = nameset - cls._facet_names
            cls._facet_names |= new_names
        for name in new_names:
            cls._add_property(name)

    def __init__(
        self,
        facets: list[Facet] | None = None,
        **kwargs: FacetValues,
    ):
        if facets is None:
            self._facets = []
        else:
            self._facets = facets
        self._init_facet_map()
        for name, values in kwargs.items():
            self[name] = values

    def _init_facet_map(self) -> None:
        self._facet_map_: dict[str, set[int]] = {}
        for i, facet in enumerate(self._facets):
            self._facet_map_.setdefault(facet.name, set())
            self._facet_map_[facet.name].add(i)

    def __getitem__(self, name: str) -> list[str]:
        if name in self._facet_names:
            return getattr(self, name)
        else:
            raise KeyError(name)

    def __setitem__(self, name: str, value: FacetValues):
        if name in self._facet_names:
            setattr(self, name, value)
        else:
            raise KeyError(name)

    def items(self) -> Iterator[tuple[str, list[str]]]:
        if not hasattr(self, "_facet_map_"):
            self._init_facet_map()
        for name in sorted(self._facet_map_.keys()):
            yield name, self[name]

    def __bool__(self) -> bool:
        return bool(self._facets)

    def _as_bytes(self) -> bytes:
        return str(tuple(self.items())).encode()

    def compute_sha(self) -> None:
        for facet in self._facets:
            if facet.sha is None:
                facet.compute_sha()
        super().compute_sha()

    def asdict(self) -> MutableMapping[str, FacetValues]:
        result: dict[str, FacetValues] = {}
        for name, facet in self.items():
            if len(facet) == 1:
                result[name] = facet[0]
            else:
                result[name] = list(facet)
        return result

    def __rich_repr__(self) -> Iterator[tuple[str, FacetValues]]:
        for name, facet in self.items():
            if len(facet) == 1:
                yield name, facet[0]
            else:
                yield name, facet

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        items = [f"{k}={v}" for k, v in self.__rich_repr__()]
        return f"{cls_name}(" + ", ".join(items) + ")"


DefaultFacets = [
    "query",
    "start",
    "end",
    "facets",
    "url",
    "data_node",
    "index_node",
    "master_id",
    "instance_id",  # search does not work with instance_id
    "title",
    "variable_long_name",
    "experiment_family",
]
BaseFacets = [
    "project",
    "mip_era",
    "experiment",
    "experiment_id",
    "institute",
    "institution_id",
    "model",
    "table_id",
    "activity_id",
    "ensemble",
    "variant_label",
    "realm",
    "frequency",
    "time_frequency",
    "variable",
    "variable_id",
    "dataset_id",
    "source_id",
    "domain",
    "driving_model",
    "rcm_name",
    "member_id",
    "cmor_table",
    "grid_label",
    "nominal_resolution",
]


Selection.reset()



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column, relationship

from esgpull.models.base import BaseNoSHA
from esgpull.models.utils import find_int, find_str

if TYPE_CHECKING:
    from esgpull.models.query import File


@dataclass
class DatasetRecord:
    dataset_id: str
    master_id: str
    version: str
    data_node: str
    size: int
    number_of_files: int

    @classmethod
    def serialize(cls, source: dict) -> DatasetRecord:
        dataset_id = find_str(source["instance_id"]).partition("|")[0]
        master_id, version = dataset_id.rsplit(".", 1)
        data_node = find_str(source["data_node"])
        size = find_int(source["size"])
        number_of_files = find_int(source["number_of_files"])
        return cls(
            dataset_id=dataset_id,
            master_id=master_id,
            version=version,
            data_node=data_node,
            size=size,
            number_of_files=number_of_files,
        )


class Dataset(BaseNoSHA):
    __tablename__ = "dataset"

    dataset_id: Mapped[str] = mapped_column(sa.String(255), primary_key=True)
    total_files: Mapped[int] = mapped_column(sa.Integer)
    created_at: Mapped[datetime] = mapped_column(
        server_default=sa.func.now(),
        default_factory=lambda: datetime.now(timezone.utc),
        init=False,
    )
    updated_at: Mapped[datetime] = mapped_column(
        server_default=sa.func.now(),
        default_factory=lambda: datetime.now(timezone.utc),
        init=False,
    )
    files: Mapped[list[File]] = relationship(
        back_populates="dataset",
        foreign_keys="[File.dataset_id]",
        primaryjoin="Dataset.dataset_id==File.dataset_id",
        default_factory=list,
        init=False,
        repr=False,
    )

    def asdict(self) -> Mapping[str, Any]:
        return {
            "dataset_id": self.dataset_id,
            "total_files": self.total_files,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
        }

    def __hash__(self) -> int:
        return hash(self.dataset_id)



import asyncio
import json

import httpx
import pytest

from esgpull.fs import FileCheck, Filesystem
from esgpull.models import File
from esgpull.processor import Task
from esgpull.result import Ok

FILE_JSON: dict = json.loads("""
{
  "id": "CMIP6.AerChemMIP.CNRM-CERFACS.CNRM-ESM2-1.hist-1950HC.r1i1p1f2.fx.sftlf.gr.v20190621.sftlf_fx_CNRM-ESM2-1_hist-1950HC_r1i1p1f2_gr.nc|esgf.ceda.ac.uk",
  "version": "1",
  "activity_drs": [
    "AerChemMIP"
  ],
  "activity_id": [
    "AerChemMIP"
  ],
  "cf_standard_name": [
    "land_area_fraction"
  ],
  "checksum": [
    "cfdc9d9113d8ab81dd51f5b133267e2d3d4db830ee0a399b444cbb1d540f512e"
  ],
  "checksum_type": [
    "SHA256"
  ],
  "citation_url": [
    "http://cera-www.dkrz.de/WDCC/meta/CMIP6/CMIP6.AerChemMIP.CNRM-CERFACS.CNRM-ESM2-1.hist-1950HC.r1i1p1f2.fx.sftlf.gr.v20190621.json"
  ],
  "data_node": "esgf.ceda.ac.uk",
  "data_specs_version": [
    "01.00.21"
  ],
  "dataset_id": "CMIP6.AerChemMIP.CNRM-CERFACS.CNRM-ESM2-1.hist-1950HC.r1i1p1f2.fx.sftlf.gr.v20190621|esgf.ceda.ac.uk",
  "dataset_id_template_": [
    "%(mip_era)s.%(activity_drs)s.%(institution_id)s.%(source_id)s.%(experiment_id)s.%(member_id)s.%(table_id)s.%(variable_id)s.%(grid_label)s"
  ],
  "directory_format_template_": [
    "%(root)s/%(mip_era)s/%(activity_drs)s/%(institution_id)s/%(source_id)s/%(experiment_id)s/%(member_id)s/%(table_id)s/%(variable_id)s/%(grid_label)s/%(version)s"
  ],
  "experiment_id": [
    "hist-1950HC"
  ],
  "experiment_title": [
    "historical forcing, but with1950s halocarbon concentrations; initialized in 1950"
  ],
  "frequency": [
    "fx"
  ],
  "further_info_url": [
    "https://furtherinfo.es-doc.org/CMIP6.CNRM-CERFACS.CNRM-ESM2-1.hist-1950HC.none.r1i1p1f2"
  ],
  "grid": [
    "data regridded to a T127 gaussian grid (128x256 latlon) from a native atmosphere T127l reduced gaussian grid"
  ],
  "grid_label": [
    "gr"
  ],
  "index_node": "esgf.ceda.ac.uk",
  "instance_id": "CMIP6.AerChemMIP.CNRM-CERFACS.CNRM-ESM2-1.hist-1950HC.r1i1p1f2.fx.sftlf.gr.v20190621.sftlf_fx_CNRM-ESM2-1_hist-1950HC_r1i1p1f2_gr.nc",
  "institution_id": [
    "CNRM-CERFACS"
  ],
  "latest": true,
  "master_id": "CMIP6.AerChemMIP.CNRM-CERFACS.CNRM-ESM2-1.hist-1950HC.r1i1p1f2.fx.sftlf.gr.sftlf_fx_CNRM-ESM2-1_hist-1950HC_r1i1p1f2_gr.nc",
  "member_id": [
    "r1i1p1f2"
  ],
  "metadata_format": "THREDDS",
  "mip_era": [
    "CMIP6"
  ],
  "model_cohort": [
    "Registered"
  ],
  "nominal_resolution": [
    "250 km"
  ],
  "pid": [
    "hdl:21.14100/37f7d29f-8d1f-354f-8554-f71adda35d22"
  ],
  "product": [
    "model-output"
  ],
  "project": [
    "CMIP6"
  ],
  "realm": [
    "atmos"
  ],
  "replica": true,
  "size": 53605,
  "source_id": [
    "CNRM-ESM2-1"
  ],
  "source_type": [
    "AOGCM",
    "BGC",
    "AER",
    "CHEM"
  ],
  "sub_experiment_id": [
    "none"
  ],
  "table_id": [
    "fx"
  ],
  "timestamp": "2020-03-10T01:53:55Z",
  "title": "sftlf_fx_CNRM-ESM2-1_hist-1950HC_r1i1p1f2_gr.nc",
  "tracking_id": [
    "hdl:21.14100/70b80649-e719-48df-9e3b-a7b7299c1c3c"
  ],
  "type": "File",
  "url": [
    "https://esgf.ceda.ac.uk/thredds/fileServer/esg_cmip6/CMIP6/AerChemMIP/CNRM-CERFACS/CNRM-ESM2-1/hist-1950HC/r1i1p1f2/fx/sftlf/gr/v20190621/sftlf_fx_CNRM-ESM2-1_hist-1950HC_r1i1p1f2_gr.nc|application/netcdf|HTTPServer",
    "https://esgf.ceda.ac.uk/thredds/dodsC/esg_cmip6/CMIP6/AerChemMIP/CNRM-CERFACS/CNRM-ESM2-1/hist-1950HC/r1i1p1f2/fx/sftlf/gr/v20190621/sftlf_fx_CNRM-ESM2-1_hist-1950HC_r1i1p1f2_gr.nc.html|application/opendap-html|OPENDAP"
  ],
  "variable": [
    "sftlf"
  ],
  "variable_id": [
    "sftlf"
  ],
  "variable_long_name": [
    "Land Area Fraction"
  ],
  "variable_units": [
    "%"
  ],
  "variant_label": [
    "r1i1p1f2"
  ],
  "retracted": false,
  "_timestamp": "2020-03-11T20:36:07.909Z",
  "score": 7.750845,
  "_version_": 1803414002300092400
}
""")


@pytest.fixture
def smallfile():
    return File.serialize(FILE_JSON)


@pytest.fixture
def fs(config):
    return Filesystem.from_config(config, install=True)


@pytest.fixture
def task(config, fs, smallfile):
    return Task(config, fs, file=smallfile)


async def run_task(task_):
    semaphore = asyncio.Semaphore(1)
    async with httpx.AsyncClient() as client:
        async for result in task_.stream(semaphore, client):
            ...
    return result


# @pytest.mark.xfail(
#     raises=(httpx.ConnectTimeout, httpx.ReadTimeout),
#     reason="this is dependent on the IPSL data node's health (unstable)",
# )
def test_task(fs, smallfile, task):
    result = asyncio.run(run_task(task))
    if not result.ok:
        raise result.err
    assert fs.finalize(smallfile) == Ok(FileCheck.Ok)
    with fs[smallfile].drs.open("rb") as f:
        data = f.read()
    assert len(data) == smallfile.size


# def test_task_url_multiple_version_correct():
#     # fmt:off
#     url_old = "http://vesg.ipsl.upmc.fr/thredds/fileServer/cmip6/CMIP/IPSL/IPSL-CM6A-LR/1pctCO2/r1i1p1f1/Oyr/bfe/gn/v20180727/bfe_Oyr_IPSL-CM6A-LR_1pctCO2_r1i1p1f1_gn_1850-1999.nc"
#     url_new = "http://vesg.ipsl.upmc.fr/thredds/fileServer/cmip6/CMIP/IPSL/IPSL-CM6A-LR/1pctCO2/r1i1p1f1/Oyr/bfe/gn/v20190305/bfe_Oyr_IPSL-CM6A-LR_1pctCO2_r1i1p1f1_gn_1850-1999.nc"
#     # fmt:on
#     download_old = Simple(auth, url=url_old)
#     download_new = Simple(auth, url=url_new)
#     assert download_old.file.filename == download_new.file.filename
#     assert download_old.file.version != download_new.file.version


# def test_task_not_esgf_url():
#     with pytest.raises(ValueError):
#         Simple(auth, url="https://www.ipsl.fr/")



"""update tables

Revision ID: 0.5.3
Revises: 0.5.2
Create Date: 2023-06-13 10:27:02.691266

"""


# revision identifiers, used by Alembic.
revision = "0.5.3"
down_revision = "0.5.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from typing import TypeVar

from esgpull.models.base import Base
from esgpull.models.dataset import Dataset, DatasetRecord
from esgpull.models.facet import Facet
from esgpull.models.file import FastFile, FileStatus
from esgpull.models.options import Option, Options
from esgpull.models.query import File, LegacyQuery, Query, QueryDict
from esgpull.models.selection import Selection
from esgpull.models.synda_file import SyndaFile
from esgpull.models.tag import Tag

Table = TypeVar("Table", bound=Base)

__all__ = [
    "Base",
    "Dataset",
    "DatasetRecord",
    "Facet",
    "FastFile",
    "File",
    "FileStatus",
    "LegacyQuery",
    "Option",
    "Options",
    "Query",
    "QueryDict",
    "Selection",
    "SyndaFile",
    "Table",
    "Tag",
]



from rich.console import Console, ConsoleOptions
from rich.measure import Measurement, measure_renderables


def short_sha(sha: str) -> str:
    return f"<{sha[:6]}>"


def rich_measure_impl(
    self,
    console: Console,
    options: ConsoleOptions,
) -> Measurement:
    renderables = list(self.__rich_console__(console, options))
    return measure_renderables(console, options, renderables)


def find_str(container: list | str) -> str:
    if isinstance(container, list):
        return find_str(container[0])
    elif isinstance(container, str):
        return container
    else:
        raise ValueError(container)


def find_int(container: list | int) -> int:
    if isinstance(container, list):
        return find_int(container[0])
    elif isinstance(container, int):
        return container
    else:
        raise ValueError(container)


def get_local_path(source: dict, version: str) -> str:
    flat_raw = {}
    for k, v in source.items():
        if isinstance(v, list) and len(v) == 1:
            flat_raw[k] = v[0]
        else:
            flat_raw[k] = v
    template = find_str(flat_raw["directory_format_template_"])
    # format: "%(a)/%(b)/%(c)/..."
    template = template.removeprefix("%(root)s/")
    template = template.replace("%(", "{")
    template = template.replace(")s", "}")
    flat_raw.pop("version", None)
    if "rcm_name" in flat_raw:  # cordex special case
        institute = flat_raw["institute"]
        rcm_name = flat_raw["rcm_name"]
        rcm_model = institute + "-" + rcm_name
        flat_raw["rcm_model"] = rcm_model
    return template.format(version=version, **flat_raw)



import click
from click.exceptions import Abort, Exit
from rich.box import MINIMAL_DOUBLE_HEAD
from rich.table import Table
from rich.text import Text

from esgpull.cli.decorators import opts
from esgpull.cli.utils import init_esgpull
from esgpull.models import sql
from esgpull.tui import Verbosity
from esgpull.utils import format_size


@click.command()
@opts.simple
@opts.all
@opts.verbosity
def status(
    simple: bool,
    all_: bool,
    verbosity: Verbosity,
):
    """
    View file queue status

    Use the `--all` flag to include already downloaded files (`done` status).
    """
    esg = init_esgpull(verbosity)
    with esg.ui.logging("status", onraise=Abort):
        status_count_size = list(esg.db.rows(sql.file.status_count_size(all_)))
        table = Table(box=MINIMAL_DOUBLE_HEAD, show_edge=False)
        table.add_column("status", justify="right", style="bold blue")
        table.add_column("files", justify="center")
        table.add_column("size", justify="right", style="magenta")
        if not status_count_size:
            esg.ui.print("Queue is empty.")
            raise Exit(0)
        if simple:
            for status, count, size in status_count_size:
                table.add_row(status.value, str(count), format_size(size))
        else:
            esg.graph.load_db()
            first_line = True
            for status, count, total_size in status_count_size:
                first_row = True
                for query in esg.graph.queries.values():
                    if not query.tracked:
                        continue
                    st_count, st_size = query.files_count_size(status)
                    if st_count:
                        if first_row:
                            if first_line:
                                first_line = False
                            else:
                                table.add_row(end_section=True)
                            table.add_row(
                                Text(status.value, justify="left"),
                                str(count),
                                format_size(total_size),
                                style="bold",
                                end_section=True,
                            )
                            first_row = False
                        table.add_row(
                            query.rich_name,
                            str(st_count),
                            format_size(st_size),
                        )
        esg.ui.print(table)



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path

import platformdirs
from typing_extensions import NotRequired, TypedDict

from esgpull.constants import INSTALLS_PATH_ENV, ROOT_ENV
from esgpull.exceptions import AlreadyInstalledName, AlreadyInstalledPath


@dataclass(init=False)
class Install:
    path: Path
    name: str | None = None

    def __init__(self, path: Path | str, name: str | None = None) -> None:
        self.path = Path(path)
        self.name = name

    def asdict(self) -> InstallDict:
        result: InstallDict = {"path": str(self.path)}
        if self.name is not None:
            result["name"] = self.name
        return result


class InstallDict(TypedDict):
    path: str
    name: NotRequired[str]


class InstallConfigDict(TypedDict):
    current: NotRequired[int]
    installs: list[InstallDict]


@dataclass(init=False)
class _InstallConfig:
    path: Path
    current_idx: int | None
    installs: list[Install]

    def __init__(self) -> None:
        self.setup()

    def setup(self, install_path: Path | None = None):
        if install_path is not None:
            user_config_dir = install_path
        elif (env := os.environ.get(INSTALLS_PATH_ENV)) is not None:
            user_config_dir = Path(env)
        else:
            user_config_dir = platformdirs.user_config_path("esgpull")
        self.path = user_config_dir / "installs.json"
        if self.path.is_file():
            with self.path.open() as f:
                content = json.load(f)
            self.current_idx = content.get("current")
            installs = content.get("installs", [])
            self.installs = [Install(**inst) for inst in installs]
        else:
            self.current_idx = None
            self.installs = []

    def fullpath(self, path: Path) -> Path:
        return path.expanduser().resolve()

    @property
    def current(self) -> Install | None:
        env = os.getenv(ROOT_ENV)
        if env is not None:
            return self.installs[int(env)]
        elif self.current_idx is not None:
            return self.installs[self.current_idx]
        else:
            return None

    @property
    def default(self) -> Path:
        return Path.home() / ".esgpull"

    def activate_msg(self, idx: int, commented: bool = False) -> str:
        install = self.installs[idx]
        name = install.name or install.path
        msg = f"""
To choose {install.path} as the current install location:

$ esgpull self choose {name}


To enable {install.path} for the current shell:

$ eval $(esgpull self activate {name})
""".strip()
        if commented:
            lines = msg.splitlines()
            return "\n".join("# " + line for line in lines)
        else:
            return msg

    def activate_needs_eval(self, idx: int) -> str:
        export = f"export {ROOT_ENV}={idx}"
        comment = self.activate_msg(idx, commented=True)
        return "\n".join([export, comment])

    def asdict(self) -> InstallConfigDict:
        result: InstallConfigDict = {
            "installs": [inst.asdict() for inst in self.installs]
        }
        if self.current_idx is not None:
            result["current"] = self.current_idx
        return result

    def add(self, path: Path, name: str | None = None) -> int:
        install = Install(self.fullpath(path), name)
        idx_path = self.index(path=install.path)
        if idx_path > -1:
            raise AlreadyInstalledPath(
                path=install.path,
                msg=self.activate_msg(idx_path),
            )
        if name is not None:
            idx_name = self.index(name=name)
            if idx_name > -1:
                raise AlreadyInstalledName(
                    name=name,
                    msg=self.activate_msg(idx_name),
                )
        self.installs.append(install)
        return len(self.installs) - 1

    def remove_current(self) -> bool:
        if self.current_idx is None:
            return False
        else:
            self.installs.pop(self.current_idx)
            self.current_idx = None
            return True

    def choose(
        self,
        *,
        idx: int | None = None,
        name: str | None = None,
        path: Path | None = None,
    ) -> None:
        if idx is not None:
            self.current_idx = idx
        elif name is not None:
            self.current_idx = self.index(name=name)
        elif path is not None:
            self.current_idx = self.index(path=path)
        else:
            self.current_idx = None
        if self.current_idx == -1:
            self.current_idx = None

    def write(self) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open("w") as f:
            json.dump(self.asdict(), f)

    def index(
        self,
        *,
        name: str | None = None,
        path: Path | None = None,
    ) -> int:
        if name is not None:
            return self._index_name(name)
        elif path is not None:
            return self._index_path(path)
        else:
            raise ValueError("nothing provided")

    def _index_name(self, name: str) -> int:
        for i, inst in enumerate(self.installs):
            if inst.name is not None and name == inst.name:
                return i
        return -1

    def _index_path(self, path: Path) -> int:
        path = self.fullpath(path)
        for i, inst in enumerate(self.installs):
            if path == inst.path:
                return i
        return -1


InstallConfig = _InstallConfig()



from __future__ import annotations

import hashlib
from collections.abc import Iterator
from dataclasses import InitVar, dataclass, field
from enum import Enum, auto
from pathlib import Path
from shutil import copyfile

import aiofiles
from aiofiles.threadpool.binary import AsyncBufferedIOBase

from esgpull.config import Config, Paths
from esgpull.models import File
from esgpull.result import Err, Ok, Result
from esgpull.tui import logger


class FileCheck(Enum):
    Missing = auto()  # not in any known paths
    Part = auto()  # {file.sha}.part exists
    BadSize = auto()  # {file.sha}.done exists AND has wrong size
    BadChecksum = auto()  # {file.sha}.done exists AND has wrong size
    Done = auto()  # {file.sha}.done exists AND is ready to be moved
    Ok = auto()  # file is in drs with everything ok

    def as_err(self, file: File) -> Exception:
        err_cls = type(str(self), (Exception,), {})
        return err_cls(file)


@dataclass
class Digest:
    file: InitVar[File]
    alg: hashlib._Hash = field(init=False)

    def __post_init__(self, file: File) -> None:
        match file.checksum_type:
            case "SHA256":
                self.alg = hashlib.sha256()
            case _:
                raise NotImplementedError

    @classmethod
    def from_path(cls, file: File, path: Path) -> Digest:
        block_size = path.stat().st_blksize
        digest = cls(file)
        with path.open("rb") as f:
            while True:
                block = f.read(block_size)
                if block == b"":
                    break
                else:
                    digest.update(block)
        return digest

    def update(self, chunk: bytes) -> None:
        self.alg.update(chunk)

    def hexdigest(self) -> str:
        return self.alg.hexdigest()


@dataclass
class Filesystem:
    paths: Paths
    disable_checksum: bool = False
    install: InitVar[bool] = True

    @staticmethod
    def from_config(config: Config, install: bool = False) -> Filesystem:
        return Filesystem(
            paths=config.paths,
            disable_checksum=config.download.disable_checksum,
            install=install,
        )

    def __post_init__(self, install: bool = True) -> None:
        if install:
            for path in self.paths.values():
                path.mkdir(parents=True, exist_ok=True)

    def __getitem__(self, file: File) -> FilePath:
        if not isinstance(file, File):
            raise TypeError(file)
        return FilePath(
            drs=self.paths.data / file.local_path / file.filename,
            tmp=self.paths.tmp / f"{file.sha}.part",
        )

    def glob_netcdf(self) -> Iterator[Path]:
        for path in self.paths.data.glob("**/*.nc"):
            yield path.relative_to(self.paths.data)

    def open(self, file: File) -> FileObject:
        return FileObject(self[file])

    def isempty(self, path: Path) -> bool:
        if next(path.iterdir(), None) is None:
            return True
        else:
            return False

    def iter_empty_parents(self, path: Path) -> Iterator[Path]:
        sample: Path | None
        for _ in range(10):  # abitrary 10 to avoid infinite loop
            sample = next(path.glob("**/*.nc"), None)
            if sample is None and self.isempty(path):
                yield path
                path = path.parent
            else:
                return

    def move_to_drs(self, file: File) -> None:
        path = self[file]
        path.drs.parent.mkdir(parents=True, exist_ok=True)
        try:
            path.done.rename(path.drs)
        except OSError as err:
            logger.error(err)
            copyfile(path.done, path.drs)
            msg = """
File rename error, shutil.copyfile was used instead.
For large files, download times might be impacted.
To address this issue, you may consider setting your `tmp` directory to the same filesystem as your `data` directory:

$ esgpull config path.tmp <some/path/on/data/filesystem>
            """.strip()
            logger.error(msg)

    def delete(self, *files: File) -> None:
        for file in files:
            path = self[file].drs
            if not path.is_file():
                continue
            path.unlink()
            logger.info(f"Deleted file {path}")
            for subpath in self.iter_empty_parents(path.parent):
                subpath.rmdir()
                logger.info(f"Deleted empty folder {subpath}")

    def compute_checksum(
        self,
        file: File,
        path: Path,
        disable_checksum: bool | None = None,
    ) -> str:
        if disable_checksum is None:
            disable_checksum = self.disable_checksum
        if disable_checksum:
            return file.checksum
        else:
            return Digest.from_path(file, path).hexdigest()

    def check_impl(
        self,
        file: File,
        path: Path,
        digest: Digest | None = None,
    ) -> FileCheck:
        if path.stat().st_size != file.size:
            return FileCheck.BadSize
        if digest is None:
            checksum = self.compute_checksum(file, path)
        else:
            checksum = digest.hexdigest()
        if checksum == file.checksum:
            return FileCheck.Ok
        else:
            return FileCheck.BadChecksum

    def check(
        self,
        file: File,
        digest: Digest | None = None,
    ) -> FileCheck:
        path = self[file]
        if path.drs.is_file():
            return self.check_impl(file, path.drs, digest)
        elif path.done.is_file():
            match self.check_impl(file, path.done, digest):
                case FileCheck.Ok:
                    return FileCheck.Done
                case check:
                    return check
        elif path.tmp.is_file():
            match self.check_impl(file, path.tmp, digest):
                case FileCheck.BadSize:
                    return FileCheck.Part
                case _:
                    raise ValueError()
        else:
            return FileCheck.Missing

    def finalize(
        self,
        file: File,
        digest: Digest | None = None,
    ) -> Result[FileCheck]:
        match self.check(file, digest=digest):
            case FileCheck.Ok:
                return Ok(FileCheck.Ok)
            case FileCheck.Done:
                self.move_to_drs(file)
                return Ok(FileCheck.Ok)
            case check:
                return Err(check, check.as_err(file))


@dataclass
class FilePath:
    drs: Path
    tmp: Path

    @property
    def done(self) -> Path:
        return self.tmp.with_suffix(".done")

    def __str__(self) -> str:
        return str(self.drs)


@dataclass
class FileObject:
    path: FilePath
    buffer: AsyncBufferedIOBase = field(init=False)

    async def __aenter__(self) -> FileObject:
        self.buffer = await aiofiles.open(self.path.tmp, "wb")
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback) -> None:
        if not self.buffer.closed:
            await self.buffer.close()

    async def write(self, chunk: bytes) -> None:
        await self.buffer.write(chunk)

    async def to_done(self) -> None:
        if not self.buffer.closed:
            await self.buffer.close()
        self.path.tmp.rename(self.path.done)



import asyncio
from pathlib import Path

import pytest

from esgpull.config import Config
from esgpull.fs import FileCheck, Filesystem
from esgpull.models import File


@pytest.fixture
def fs_no_install():
    config = Config()
    return Filesystem.from_config(config, install=False)


@pytest.fixture
def fs() -> Filesystem:
    config = Config()
    return Filesystem.from_config(config, install=True)


def test_install(root: Path, fs: Filesystem):
    assert str(fs.paths.data) == str(root / "data")
    assert str(fs.paths.db) == str(root / "db")
    assert str(fs.paths.log) == str(root / "log")
    assert str(fs.paths.tmp) == str(root / "tmp")
    assert fs.paths.data.is_dir()
    assert fs.paths.db.is_dir()
    assert fs.paths.log.is_dir()
    assert fs.paths.tmp.is_dir()


def test_no_install(root: Path, fs_no_install: Filesystem):
    assert str(fs_no_install.paths.data) == str(root / "data")
    assert str(fs_no_install.paths.db) == str(root / "db")
    assert str(fs_no_install.paths.log) == str(root / "log")
    assert str(fs_no_install.paths.tmp) == str(root / "tmp")
    assert not fs_no_install.paths.data.is_dir()
    assert not fs_no_install.paths.db.is_dir()
    assert not fs_no_install.paths.log.is_dir()
    assert not fs_no_install.paths.tmp.is_dir()


def test_file_paths(fs: Filesystem, file: File):
    file.sha = "1234"
    path = fs[file]
    assert path.drs == fs.paths.data / "project/folder/file.nc"
    assert path.tmp == fs.paths.tmp / "1234.part"


async def write_steps(fs: Filesystem, file: File):
    async with fs.open(file) as f:
        await f.write(b"")


def test_write(fs: Filesystem, file: File):
    asyncio.run(write_steps(fs, file))
    for path in fs.glob_netcdf():
        assert str(path) == "project/folder/file.nc"


@pytest.mark.parametrize(
    "expected_check,content,kind,size,checksum",
    [
        pytest.param(
            FileCheck.Missing,
            None,
            None,
            None,
            None,
            id="file does not exist anywhere",
        ),
        pytest.param(
            FileCheck.Part,
            "cont",
            "tmp",
            7,
            None,
            id="size not matching content (tmp)",
        ),
        pytest.param(
            FileCheck.BadSize,
            "content",
            "done",
            0,
            None,
            id="size not matching content (done)",
        ),
        pytest.param(
            FileCheck.BadSize,
            "content",
            "drs",
            0,
            None,
            id="size not matching content (drs)",
        ),
        pytest.param(
            FileCheck.BadChecksum,
            "content",
            "tmp",
            7,
            "",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="checksum not matching content (tmp)",
        ),
        pytest.param(
            FileCheck.BadChecksum,
            "content",
            "done",
            7,
            "0",
            id="checksum not matching content (done)",
        ),
        pytest.param(
            FileCheck.BadChecksum,
            "content",
            "drs",
            7,
            "0",
            id="checksum not matching content (drs)",
        ),
        pytest.param(
            FileCheck.Done,
            "content",
            "tmp",
            7,
            "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f73",
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
            id="size and checksum match content (tmp)",
        ),
        pytest.param(
            FileCheck.Done,
            "content",
            "done",
            7,
            "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f73",
            id="size and checksum match content (done)",
        ),
        pytest.param(
            FileCheck.Ok,
            "content",
            "drs",
            7,
            "ed7002b439e9ac845f22357d822bac1444730fbdb6016d3ec9432297b9ec9f73",
            id="size and checksum match content (drs)",
        ),
    ],
)
def test_check(
    root: Path,
    fs: Filesystem,
    file: File,
    expected_check: FileCheck,
    content: str | None,
    kind: str,
    size: int,
    checksum: str,
):
    if content is not None:
        file.size = size
        file.checksum = checksum
        file.checksum_type = "SHA256"
        if kind == "tmp":
            path = fs[file].tmp
        elif kind == "done":
            path = fs[file].done
        elif kind == "drs":
            path = fs[file].drs
            path.parent.mkdir(parents=True)
        else:
            path = root / kind
        with path.open("wb") as f:
            f.write(str(content).encode())
    check = fs.check(file)
    assert check == expected_check



from pathlib import Path
from time import perf_counter

from click.testing import CliRunner

from esgpull import Esgpull
from esgpull.cli.add import add
from esgpull.cli.config import config as config_cmd
from esgpull.cli.update import update
from esgpull.config import Config


def test_fast_update(config: Config):
    config.generate(overwrite=True)
    runner = CliRunner()
    result_config = runner.invoke(config_cmd, ["api.page_limit", "10000"])
    assert result_config.exit_code == 0
    result_add = runner.invoke(
        add,
        [
            "table_id:fx",
            "experiment_id:dcpp*",
            "--distrib",
            "false",
            "--track",
        ],
    )
    assert result_add.exit_code == 0
    start = perf_counter()
    result_update = runner.invoke(update, ["--yes"])
    stop = perf_counter()
    assert result_update.exit_code == 0
    assert stop - start < 60  # 60 seconds to fetch ~4k files is plenty enough


def test_update_updates_timestamp(root: Path, config: Config):
    config.generate(overwrite=True)
    runner = CliRunner()
    result_config = runner.invoke(config_cmd, ["api.page_limit", "10000"])
    assert result_config.exit_code == 0

    # Add a small query - using a small dataset to keep test fast
    result_add = runner.invoke(
        add,
        [
            "project:CMIP6",
            "table_id:fx",
            "variable_id:areacella",
            "experiment_id:1pctCO2",
            "--distrib",
            "true",
            "--track",
        ],
    )
    assert result_add.exit_code == 0

    # Initialize Esgpull to get initial timestamp
    esg = Esgpull(root)
    esg.graph.load_db()

    # Get the only query in the database
    assert len(esg.graph._shas) == 1, "Expected only one query in database"
    query_id = list(esg.graph._shas)[0]
    query = esg.graph.get(query_id)

    initial_timestamp = query.updated_at

    # Run update with --yes to automatically add files
    result_update = runner.invoke(update, [query_id, "--yes"])
    assert result_update.exit_code == 0

    # Get the query again and check timestamp
    esg = Esgpull(root)  # Reinitialize to ensure fresh data
    esg.graph.load_db()
    query = esg.graph.get(query_id)

    # Verify the timestamp was updated
    assert query.updated_at > initial_timestamp, (
        "updated_at timestamp was not updated after adding files"
    )

    # Save the timestamp after files were added
    timestamp_after_update = query.updated_at

    # Run a second update - no new files should be added
    result_update_2 = runner.invoke(update, [query_id, "--yes"])
    assert result_update_2.exit_code == 0

    # Get the query again
    esg = Esgpull(root)  # Reinitialize to ensure fresh data
    esg.graph.load_db()
    query = esg.graph.get(query_id)

    # Verify the timestamp was NOT updated when no new files were added
    assert query.updated_at == timestamp_after_update, (
        "updated_at timestamp should not change when no new files are added"
    )



# from __future__ import annotations

# import click
# from click.exceptions import Abort, Exit

# from esgpull import Esgpull
# from esgpull.cli.decorators import args, opts
# from esgpull.cli.utils import load_facets
# from esgpull.db.models import File
# from esgpull.tui import Verbosity
# from esgpull.utils import format_size


# @click.command()
# @opts.distrib
# @opts.dry_run
# @opts.force
# @opts.replica
# @opts.since
# @opts.selection_file
# @opts.verbosity
# @args.facets
# def install(
#     facets: list[str],
#     distrib: bool,
#     dry_run: bool,
#     force: bool,
#     replica: bool | None,
#     selection_file: str | None,
#     since: str | None,
#     verbosity: Verbosity,
# ) -> None:
#     esg = Esgpull(verbosity=verbosity)
#     with (
#         esg.context(
#             distrib=distrib,
#             latest=True,
#             since=since,
#             replica=replica,
#         ) as ctx,
#         esg.ui.logging("install", onraise=Abort),
#     ):
#         load_facets(ctx.query, facets, selection_file)
#         if not ctx.query.dump():
#             raise click.UsageError("No search terms provided.")
#         hits = ctx.file_hits
#         nb_files = sum(hits)
#         esg.ui.print(f"Found {nb_files} files.")
#         if nb_files > 500 and distrib:
#             # Enable better distrib
#             ctx.index_nodes = esg.fetch_index_nodes()
#         if dry_run:
#             queries = ctx._build_queries_search(
#                 hits, file=True, max_results=nb_files, offset=0
#             )
#             esg.ui.print(queries)
#             raise Exit(0)
#         if not force and nb_files > 5000:
#             nb_req = nb_files // esg.config.api.page_limit
#             message = f"{nb_req} requests will be send to ESGF. Continue?"
#             if not esg.ui.ask(message, default=True):
#                 raise Abort
#         results = ctx.search(
#             file=True,
#             max_results=None,
#             offset=0,
#             hits=hits,
#         )
#         files = [File.from_dict(result) for result in results]
#         total_size = sum([file.size for file in files])
#         esg.ui.print(f"Total size: {format_size(total_size)}")
#         if not force:
#             if not esg.ui.ask("Continue?", default=True):
#                 raise Abort
#         to_download, already_on_disk = esg.install(*files)
#         if to_download:
#             nb = len(to_download)
#             s = "s" if nb > 1 else ""
#             esg.ui.print(f"Installed {nb} new file{s} ready for download.")
#         if already_on_disk:
#             nb = len(already_on_disk)
#             s = "s" if nb > 1 else ""
#             esg.ui.print(f"Tracking {nb} file{s} already downloaded.")
#         if not to_download and not already_on_disk:
#             esg.ui.print("All files are already in the database.")



from __future__ import annotations

import click
from click.exceptions import Abort, Exit
from rich.box import MINIMAL_DOUBLE_HEAD
from rich.table import Table
from rich.text import Text

from esgpull.cli.decorators import args, opts
from esgpull.cli.utils import init_esgpull, valid_name_tag
from esgpull.tui import Verbosity


@click.command()
@args.query_ids
@opts.record
@opts.verbosity
def track(
    query_ids: tuple[str],
    record: bool,
    verbosity: Verbosity,
) -> None:
    """
    Track queries

    As a side effect, tracking a query applies all default options to the query,
    so that modifications of the config's default options have no impact on
    previouly tracked queries.
    """
    esg = init_esgpull(verbosity, record=record)
    with esg.ui.logging("track", onraise=Abort):
        for sha in query_ids:
            if not valid_name_tag(esg.graph, esg.ui, sha, None):
                esg.ui.raise_maybe_record(Exit(1))
            query = esg.graph.get(sha)
            if query.tracked:
                esg.ui.print(f"{query.rich_name} is already tracked.")
                esg.ui.raise_maybe_record(Exit(0))
            if esg.graph.get_children(query.sha):
                msg = f"{query.rich_name} has children, track anyway?"
                if not esg.ui.ask(msg, default=False):
                    esg.ui.raise_maybe_record(Abort)
            expanded = esg.graph.expand(query.sha)
            tracked_query = query.clone(compute_sha=False)
            tracked_query.track(expanded.options)
            if query.sha != tracked_query.sha:
                msg = f"For {query.rich_name} to become tracked, options must be set."
                esg.ui.print(msg)
                table = Table(
                    box=MINIMAL_DOUBLE_HEAD,
                    show_edge=False,
                    show_lines=True,
                )
                table.add_column(Text("before", justify="center"))
                table.add_column(Text("after", justify="center"))
                table.add_row(query, tracked_query)
                esg.ui.print(table)
                if not esg.ui.ask("Apply changes?"):
                    esg.ui.raise_maybe_record(Abort)
            esg.graph.replace(query, tracked_query)
            esg.graph.merge()
            esg.ui.print(f":+1: {tracked_query.rich_name} is now tracked.")
        esg.ui.raise_maybe_record(Exit(0))


@click.command()
@args.query_ids
@opts.verbosity
def untrack(
    query_ids: tuple[str],
    verbosity: Verbosity,
) -> None:
    """
    Untrack queries
    """
    esg = init_esgpull(verbosity)
    with esg.ui.logging("untrack", onraise=Abort):
        for sha in query_ids:
            if not valid_name_tag(esg.graph, esg.ui, sha, None):
                raise Exit(1)
            query = esg.graph.get(sha)
            if not query.tracked:
                esg.ui.print(f"Query {query.rich_name} is already untracked.")
                raise Exit(0)
            query.untrack()
            esg.graph.merge()
            esg.ui.print(f":+1: Query {query.rich_name} is no longer tracked.")



from __future__ import annotations

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column

from esgpull.models.base import Base


class Tag(Base):
    __tablename__ = "tag"

    name: Mapped[str] = mapped_column(sa.String(255))
    description: Mapped[str | None] = mapped_column(sa.Text, default=None)

    def _as_bytes(self) -> bytes:
        return self.name.encode()

    def __hash__(self) -> int:
        return hash(self._as_bytes())



from __future__ import annotations

import atexit
import logging
import sys
from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from datetime import datetime
from enum import IntEnum
from json import dumps as json_dumps
from pathlib import Path
from typing import Any, TypeVar

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup

import click.exceptions
from attrs import define, field
from rich.console import Console, Group, RenderableType
from rich.live import Live
from rich.logging import RichHandler
from rich.progress import Progress, ProgressColumn, track
from rich.prompt import Confirm, Prompt
from rich.status import Status
from rich.syntax import Syntax
from rich.text import Text
from tomlkit import dumps as tomlkit_dumps
from yaml import dump as yaml_dump

from esgpull.config import Config
from esgpull.constants import ESGPULL_DEBUG, ESGPULL_DEBUG_LOCALS

logger = logging.getLogger("esgpull")
logging.root.setLevel(logging.DEBUG)
logging.root.addHandler(logging.NullHandler())

# sqlalchemy is very verbose
logging.getLogger("sqlalchemy.engine").setLevel(logging.WARNING)

_console = Console(highlight=True)
_err_console = Console(stderr=True)
_record_console = Console(highlight=True, record=True)

T = TypeVar("T")


class Verbosity(IntEnum):
    Normal = 0
    Errors = 1
    Detail = 2
    Debug = 3

    def get_level(self) -> int:
        levels = [
            logging.WARNING,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
        ]
        return levels[self]

    def render(self) -> Text:
        return Text(self.name.upper(), style=f"logging.level.{self.name}")


class DummyConsole:
    def print(self, msg: str) -> None:
        pass


class DummyLive:
    def __enter__(self) -> DummyLive:
        return self

    def __exit__(self, *args): ...

    @property
    def console(self) -> DummyConsole:
        return DummyConsole()


def yaml_syntax(data: Mapping[str, Any]) -> Syntax:
    return Syntax(yaml_dump(data, sort_keys=False), "yaml", theme="ansi_dark")


def toml_syntax(data: Mapping[str, Any]) -> Syntax:
    return Syntax(tomlkit_dumps(data), "toml", theme="ansi_dark")


LOG_FORMAT = "[%(asctime)s]  %(levelname)-10s%(name)s\n%(message)s\n"
LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
FILE_DATE_FORMAT = "%Y-%m-%d_%H-%M-%S"


class ExceptionToDebugFilter(logging.Filter):
    def filter(self, record):
        # If it's an exception record, change its level to DEBUG
        if record.exc_info is not None:
            record.levelno = logging.DEBUG
            record.levelname = "DEBUG"
        return True


class ErrorCountColumn(ProgressColumn):
    def render(self, task):
        nb_errors = task.fields.get("nb_errors", 0)

        if nb_errors == 0:
            return Text("")

        suffix = "" if nb_errors == 1 else "s"
        return Text(f"({nb_errors} download{suffix} failed)", style="red")


@define
class UI:
    path: Path = field(converter=Path)
    verbosity: Verbosity = Verbosity.Normal
    logfile: bool = True
    record: bool = False
    default_onraise: type[Exception] | Exception | None = None
    # max_size: int = 1 << 30

    @property
    def console(self) -> Console:
        if self.record:
            return _record_console
        else:
            return _console

    @property
    def err_console(self) -> Console:
        if self.record:
            return _record_console
        else:
            return _err_console

    @staticmethod
    def from_config(
        config: Config,
        verbosity: Verbosity = Verbosity.Normal,
        record: bool = False,
    ) -> UI:
        return UI(config.paths.log, verbosity=verbosity, record=record)

    @contextmanager
    def logging(
        self,
        modulename: str = "",
        onraise: type[Exception] | Exception | None = None,
        record: bool | None = None,
    ):
        if self.verbosity > Verbosity.Normal:
            logger.setLevel(logging.INFO)
        if record is not None:
            self.record = record
        handler: logging.Handler
        temp_path: Path | None = None
        fmt = LOG_FORMAT
        datefmt = LOG_DATE_FORMAT
        if self.verbosity >= Verbosity.Errors:
            if self.err_console.is_terminal or self.err_console.is_jupyter:
                handler = RichHandler(
                    console=self.err_console,
                    show_path=False,
                    markup=True,
                )
                fmt = "[yellow]· %(name)s ·[/]\n%(message)s"
                datefmt = "[%X]"
            else:
                handler = logging.StreamHandler()
            handler.setLevel(self.verbosity.get_level())
        elif self.logfile:
            date = datetime.utcnow().strftime(FILE_DATE_FORMAT)
            filename = "-".join(["esgpull", modulename, date]) + ".log"
            temp_path = self.path / filename
            handler = logging.FileHandler(temp_path)
            handler.setLevel(logging.DEBUG)
            error_print_handler = logging.StreamHandler()
            error_print_handler.setLevel(logging.ERROR)
            error_print_handler.setFormatter(
                logging.Formatter(fmt=fmt, datefmt=datefmt)
            )
            logging.root.addHandler(error_print_handler)
        else:
            handler = logging.NullHandler()
        handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
        logger.addFilter(ExceptionToDebugFilter())
        logging.root.addHandler(handler)
        try:
            yield
        except (click.exceptions.Exit, click.exceptions.Abort):
            if temp_path is not None:
                atexit.register(lambda: temp_path.unlink(missing_ok=True))
            raise
        except click.exceptions.ClickException:
            raise
        except BaseException as exc:
            tb = exc.__traceback__
            while True:
                if tb is None:
                    break
                elif tb.tb_next is None:
                    break
                tb = tb.tb_next
            if tb is None:
                f_locals = {}
            else:
                f_locals = tb.tb_frame.f_locals
            locals_text = self.render(f_locals, highlight=False)
            logging.root.debug(f"Locals:\n{locals_text}")
            logging.root.exception("")
            self.print(f"[red]{type(exc).__name__}[/]: {exc}", err=True)
            if self.verbosity < Verbosity.Errors and self.logfile:
                self.print(
                    f"See [yellow]{temp_path}[/] for error log.",
                    err=True,
                )
            if ESGPULL_DEBUG or ESGPULL_DEBUG_LOCALS:
                from rich.traceback import install

                install(show_locals=ESGPULL_DEBUG_LOCALS)
                raise
            elif onraise is not None:
                raise onraise
            elif self.default_onraise is not None:
                raise self.default_onraise
            else:
                raise
        else:
            if temp_path is not None:
                atexit.register(lambda: temp_path.unlink(missing_ok=True))
        finally:
            logging.root.removeHandler(handler)

    def print(
        self,
        msg: Any,
        err: bool = False,
        json: bool = False,
        yaml: bool = False,
        toml: bool = False,
        verbosity: Verbosity = Verbosity.Normal,
        **kwargs: Any,
    ) -> None:
        if self.verbosity >= verbosity:
            console = self.err_console if err else self.console
            if json:
                console.print_json(json_dumps(msg), **kwargs)
            elif yaml:
                console.print(yaml_syntax(msg), **kwargs)
            elif toml:
                console.print(toml_syntax(msg), **kwargs)
            else:
                if not console.is_interactive:
                    kwargs.setdefault("crop", False)
                    kwargs.setdefault("overflow", "ignore")
                console.print(msg, **kwargs)

    def render(
        self,
        msg: Any,
        json: bool = False,
        yaml: bool = False,
        toml: bool = False,
        **kwargs: Any,
    ) -> str:
        with self.console.capture() as capture:
            if json:
                self.console.print_json(json_dumps(msg), **kwargs)
            elif yaml:
                self.console.print(yaml_syntax(msg), **kwargs)
            elif toml:
                self.console.print(toml_syntax(msg), **kwargs)
            else:
                self.console.print(msg, **kwargs)
        return capture.get()

    def live(
        self,
        first: RenderableType,
        *rest: RenderableType,
        disable: bool = False,
    ) -> Live | DummyLive:
        if disable:
            return DummyLive()
        if not rest:
            renderables = first
        else:
            renderables = Group(first, *rest)
        # use _console to avoid recording the progress bar
        return Live(renderables, console=_console)

    def track(self, iterable: Iterable[T], **kwargs) -> Iterable[T]:
        # use _console to avoid recording the progress bar
        return track(iterable, console=_console, **kwargs)

    def make_progress(
        self,
        *columns: str | ProgressColumn,
        **kwargs: Any,
    ) -> Progress:
        # use _console to avoid recording the progress bar
        return Progress(*columns, console=_console, **kwargs)

    def spinner(self, msg: str) -> Status:
        # use _console to avoid recording the spinner
        return _console.status(msg, spinner="earth")

    def ask(self, msg: str, default: bool | None = None) -> bool:
        if default is not None:
            return Confirm.ask(msg, default=default, console=self.console)
        else:
            return Confirm.ask(msg, console=self.console)

    def choice(
        self,
        msg: str,
        choices: list[str],
        default: str | None = None,
        show_choices: bool = True,
    ) -> str:
        if default is not None:
            return Prompt.ask(
                msg,
                choices=choices,
                default=default,
                show_choices=show_choices,
                console=self.console,
            )
        else:
            return Prompt.ask(
                msg,
                choices=choices,
                show_choices=show_choices,
                console=self.console,
            )

    def prompt(
        self, msg: str, default: str | None = None, password: bool = False
    ) -> str:
        if default is not None:
            return Prompt.ask(
                msg,
                default=default,
                password=password,
                console=self.console,
            )
        else:
            return Prompt.ask(msg, password=password, console=self.console)

    def rule(self, msg: str):
        self.console.rule(msg)

    def export_svg(self) -> Path:
        date = datetime.utcnow().strftime(FILE_DATE_FORMAT)
        filename = "-".join(["record", date]) + ".svg"
        output_path = self.path / filename
        with output_path.open("w") as f:
            f.write(_record_console.export_svg())
        return output_path

    def raise_maybe_record(
        self,
        exc: type[Exception] | Exception | BaseExceptionGroup,
    ) -> None:
        if self.record:
            output_path = self.export_svg()
            self.print(f":+1: Console output exported to {output_path}")
        raise exc


TempUI = UI(
    "/tmp",
    Verbosity.Errors,
    logfile=False,
    default_onraise=click.exceptions.Exit(1),
)



import asyncio
import datetime
from typing import Callable, Coroutine, TypeVar

from rich.filesize import _to_str

T = TypeVar("T")


def sync(
    coro: Coroutine[None, None, T],
    before_cb: Callable | None = None,
    after_cb: Callable | None = None,
) -> T:
    if before_cb is not None:
        before_cb()
    result = asyncio.run(coro)
    if after_cb is not None:
        after_cb()
    return result


def format_size(size: int) -> str:
    return _to_str(
        size,
        ("kiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"),
        1024,
        precision=1,
        separator=" ",
    )


def parse_date(
    date: str | datetime.datetime, fmt: str = "%Y-%m-%d"
) -> datetime.datetime:
    match date:
        case datetime.datetime():
            ...
        case str():
            date = datetime.datetime.strptime(date, fmt)
        case _:
            raise ValueError(date)
    return date


def format_date(date: str | datetime.datetime, fmt: str = "%Y-%m-%d") -> str:
    return parse_date(date, fmt).strftime(fmt)


def format_date_iso(
    date: str | datetime.datetime, fmt: str = "%Y-%m-%d"
) -> str:
    return parse_date(date, fmt).replace(microsecond=0).isoformat() + "Z"



"""add_dataset_tracking

Revision ID: e7edab5d4e4b
Revises: 0.8.0
Create Date: 2025-05-23 17:38:22.066153

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e7edab5d4e4b'
down_revision = '0.8.0'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('dataset',
    sa.Column('dataset_id', sa.String(length=255), nullable=False),
    sa.Column('total_files', sa.Integer(), nullable=False),
    sa.Column('created_at', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.Column('updated_at', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.PrimaryKeyConstraint('dataset_id')
    )
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.create_foreign_key('fk_file_dataset', 'dataset', ['dataset_id'], ['dataset_id'])

    # ### end Alembic commands ###
    
def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('file', schema=None) as batch_op:
        batch_op.drop_constraint('fk_file_dataset', type_='foreignkey')

    op.drop_table('dataset')
    # ### end Alembic commands ###
